# coding: utf-8
import uuid
import functools

from django.db.models import Q, CharField, TextField, ForeignKey, BooleanField
from django.contrib.gis.db.models import PointField
from django.contrib.gis.measure import D
from django.db.models.fields.related import ManyToManyField
from django.contrib.auth import get_user_model
from rest_framework.exceptions import ValidationError
from rest_framework.filters import BaseFilterBackend
from concrete_datastore.api.v1.datetime import format_datetime, ensure_pendulum
from django.contrib.gis.geos import Point

# target_field type is date or datetime or int or float
RANGEABLE_TYPES = (
    'DateTimeField',
    'DateField',
    'DecimalField',
    'IntegerField',
    'FloatField',
)


def convert_type(string, field_type, close_period=True):
    if string == '':
        raise ValidationError(
            {
                "message": "Attempting to convert an empty string to a date format"
            }
        )
    if field_type in ('DateTimeField', 'DateField'):
        if field_type == 'DateTimeField':
            # Expected format YYYY-MM-DD HH:MM[:ss[.uuuuuu]][TZ]
            dt = ensure_pendulum(string)
            if close_period:
                dt.end_of('day')
            else:
                dt.start_of('day')
            return format_datetime(dt)
        if field_type == 'DateField':
            # Expected YYYY-MM-DD
            dt = ensure_pendulum(string)
            # Deactivated by lco, a date is a date, no time in it
            # if close_period:
            #     dt.end_of('day')
            return dt.to_date_string()
    elif field_type in ('DecimalField', 'FloatField'):
        return float(string)
    else:
        return int(string)


class CustomShemaOperationParameters:
    def return_if_not_details(self, view, value, extra_condition=True):
        if view.detail is False and extra_condition:
            return value
        return []


class FilterDistanceBackend(BaseFilterBackend, CustomShemaOperationParameters):
    def remove_from_queryset(self, view):
        #: Remove PointField field from filterset_fields because
        #: they cannot be filtered with the other filter backends
        filterset_fields = getattr(view, 'filterset_fields', ())
        new_filterset_fields = tuple(
            filter_field
            for filter_field in filterset_fields
            if (
                view.model_class._meta.get_field(
                    filter_field
                ).get_internal_type()
                != 'PointField'
            )
        )
        setattr(view, 'filterset_fields', new_filterset_fields)

    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__distance',
                'required': False,
                'in': 'query',
                'description': 'DISTANCE,LONGITUDE,LATITUDE',
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if view.model_class._meta.get_field(field_name).get_internal_type()
            == 'PointField'
        ]
        self.remove_from_queryset(view=view)
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        # Only applicable on PointField objects
        #: The filter is __distance=XXX,LONGITUDE,LATITUDE
        q_object = None
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())
        for param in query_params:
            if not param.endswith('__distance'):
                continue
            param_field = param.replace('__distance', '')
            if param_field not in filterset_fields:
                continue
            if (
                queryset.model._meta.get_field(param_field).get_internal_type()
                != 'PointField'
            ):
                continue
            #: The split should have three elements
            values = query_params.get(param).split(',')
            if len(values) != 3:
                raise ValidationError(
                    "Distance filter needs the following parameters: "
                    "Distance, longitude and latitude."
                )
            distance = int(values[0])
            longitude = float(values[1])
            latitude = float(values[2])

            point = Point(longitude, latitude)
            custom_filter = {
                '{}__distance_lte'.format(param_field): (point, D(m=distance))
            }
            if q_object is None:
                q_object = Q(**custom_filter)
            else:
                q_object &= Q(**custom_filter)

        self.remove_from_queryset(view=view)
        if q_object is None:
            return queryset

        return queryset.filter(q_object)


class FilterUserByLevel(BaseFilterBackend, CustomShemaOperationParameters):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': 'level',
                'required': False,
                'in': 'query',
                'schema': {'type': 'string'},
            },
            {
                'name': 'atleast',
                'required': False,
                'in': 'query',
                'schema': {'type': 'string'},
            },
        ]
        return self.return_if_not_details(
            view=view,
            value=params,
            extra_condition=view.model_class == get_user_model(),
        )

    def filter_queryset(self, request, queryset, view):
        # Only appyable on user models
        if view.basename.lower() != 'user':
            return queryset

        query_params = request.query_params
        if 'level' in query_params:
            return queryset.model.filter_by_exact_level(
                queryset=queryset, level=query_params.get('level')
            )
        if 'atleast' in query_params:
            return queryset.model.filter_by_at_least_level(
                queryset=queryset, level=query_params.get('atleast')
            )

        return queryset


class FilterSupportingOrBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__in',
                'required': False,
                'in': 'query',
                'description': 'List of values separated by comma',
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if type(view.model_class._meta.get_field(field_name))
            != ManyToManyField
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())

        q_object = None
        for param in query_params:
            if not param.endswith('__in'):
                continue
            if param.replace('__in', '') not in filterset_fields:
                continue

            values = query_params.get(param).split(',')
            if isinstance(
                queryset.model._meta.get_field(param.replace('__in', '')),
                BooleanField,
            ):
                if set(values).difference(['True', 'False', 'None']):
                    raise ValidationError(
                        {
                            "message": (
                                f"{param}: {values} must contain olny 'True', "
                                "'False' and/or 'None' (case sensitive)"
                            )
                        }
                    )
            custom_filter = {param: values}
            if q_object is None:
                q_object = Q(**custom_filter)
            else:
                q_object &= Q(**custom_filter)
        if q_object is None:
            return queryset

        return queryset.filter(q_object)


class FilterSupportingContainsBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__contains',
                'required': False,
                'in': 'query',
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if isinstance(
                view.model_class._meta.get_field(field_name),
                (CharField, TextField),
            )
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())

        q_object = None
        for param in query_params:
            if not param.endswith('__contains'):
                continue
            param_field = param.replace('__contains', '')
            if param_field not in filterset_fields:
                continue
            if not isinstance(
                queryset.model._meta.get_field(param_field),
                (CharField, TextField),
            ):
                continue

            custom_filter = {param: query_params.get(param)}
            if q_object is None:
                q_object = Q(**custom_filter)
            else:
                q_object &= Q(**custom_filter)
        if q_object is None:
            return queryset

        return queryset.filter(q_object)


class FilterSupportingEmptyBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__isempty',
                'required': False,
                'in': 'query',
                'description': 'True or False',
                'schema': {'type': 'boolean'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if isinstance(
                view.model_class._meta.get_field(field_name),
                (CharField, TextField),
            )
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())

        q_object = None
        for param in query_params:
            if not param.endswith('__isempty'):
                continue
            if query_params.get(param) != 'true':
                continue
            param = param.replace('__isempty', '')
            if param not in filterset_fields:
                continue
            if isinstance(
                queryset.model._meta.get_field(param), (CharField, TextField)
            ):
                custom_filter = {'{}__exact'.format(param): ''}
            else:
                continue

            if q_object is None:
                q_object = Q(**custom_filter)
            else:
                q_object &= Q(**custom_filter)
        if q_object is None:
            return queryset

        return queryset.filter(q_object)


class FilterSupportingRangeBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__range',
                'required': False,
                'in': 'query',
                'description': 'A range of values separated by comma',
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            + ('creation_date', 'modification_date')
            if view.model_class._meta.get_field(field_name).get_internal_type()
            in RANGEABLE_TYPES
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ()) + (
            'creation_date',
            'modification_date',
        )

        q_object = None
        for param in query_params:
            if not param.endswith('__range'):
                continue
            target_field = param.replace('__range', '')
            if target_field not in filterset_fields:
                continue

            target_field_type = queryset.model._meta.get_field(
                target_field
            ).get_internal_type()
            if target_field_type not in RANGEABLE_TYPES:
                continue

            values = query_params.get(param).split(',')
            correct_range_values = len(values) == 2

            if not correct_range_values:
                continue

            range_start, range_end = values

            if range_start == '' and range_end == '':
                continue

            elif range_start != '' and range_end == '':
                param = param.replace('__range', '__gte')
                values = convert_type(
                    range_start, target_field_type, close_period=False
                )

            elif range_start == '' and range_end != '':
                param = param.replace('__range', '__lte')
                values = convert_type(
                    range_end, target_field_type, close_period=True
                )

            else:
                values = (
                    convert_type(
                        range_start, target_field_type, close_period=False
                    ),
                    convert_type(
                        range_end, target_field_type, close_period=True
                    ),
                )

            custom_filter = {param: values}
            if q_object is None:
                q_object = Q(**custom_filter)
            else:
                q_object &= Q(**custom_filter)

        if q_object is None:
            return queryset

        return queryset.filter(q_object)


class FilterSupportingComparaisonBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        suffixes_map = {
            'gte': 'get the values greater than or equal to a given value',
            'lte': 'get the values less than or equal to a given value',
            'gt': 'get the values greater than to a given value',
            'lt': 'get the values less than to a given value',
        }
        params = []
        for key, description in suffixes_map.items():
            params.extend(
                [
                    {
                        'name': f'{field_name}__{key}',
                        'required': False,
                        'in': 'query',
                        'description': description,
                        'schema': {'type': 'string'},
                    }
                    for field_name in getattr(view, 'filterset_fields', ())
                    + ('creation_date', 'modification_date')
                    if view.model_class._meta.get_field(
                        field_name
                    ).get_internal_type()
                    in RANGEABLE_TYPES
                ]
            )
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ()) + (
            'creation_date',
            'modification_date',
        )

        def get_param_from_query(param):
            target_field = param
            if target_field.endswith('__gte'):
                return target_field.replace('__gte', '')
            if target_field.endswith('__lte'):
                return target_field.replace('__lte', '')
            if target_field.endswith('__gt'):
                return target_field.replace('__gt', '')
            if target_field.endswith('__lt'):
                return target_field.replace('__lt', '')
            return None

        q_object = None

        def get_custom_filters(param):
            target_field = get_param_from_query(param)
            if target_field not in filterset_fields:
                return None

            target_field_type = queryset.model._meta.get_field(
                target_field
            ).get_internal_type()
            if target_field_type not in RANGEABLE_TYPES:
                return None
            value = convert_type(query_params.get(param), target_field_type)
            if value is None:
                return None
            return {param: value}

        filters = [
            get_custom_filters(qp)
            for qp in query_params
            if get_custom_filters(qp) is not None
        ]
        if len(filters) == 0:
            return queryset
        q_object = functools.reduce(lambda a, b: a & Q(**b), filters, Q())

        return queryset.filter(q_object)


class FilterSupportingForeignKey(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}_uid',
                'required': False,
                'in': 'query',
                'description': 'UID of the FK',
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if type(view.model_class._meta.get_field(field_name)) == ForeignKey
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())
        q_object = None
        for param in query_params:
            if not param.endswith('_uid'):
                continue
            cleaned_param = param.replace('_uid', '')
            if cleaned_param not in filterset_fields:
                continue
            cleaned_param_type = queryset.model._meta.get_field(cleaned_param)
            if not type(cleaned_param_type) == ForeignKey:
                continue
            value = query_params.get(param)
            custom_filter = {cleaned_param: value}
            #:  "value" must be a valid UUID4, otherwise raise ValidationError
            try:
                #:  raises ValueError if not UUID4
                uuid.UUID(value, version=4)
            except ValueError:
                raise ValidationError(
                    {"message": f'{param}: « {value} » is not a valid UUID'}
                )

            if q_object is None:
                q_object = Q(**custom_filter)
            else:
                q_object &= Q(**custom_filter)
        if q_object is None:
            return queryset

        return queryset.filter(q_object)


class FilterForeignKeyIsNullBackend(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__isull',
                'required': False,
                'in': 'query',
                'schema': {'type': 'boolean'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if type(view.model_class._meta.get_field(field_name))
            in (ForeignKey, ManyToManyField)
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())

        q_object = None
        for param in query_params:
            if not param.endswith('__isnull'):
                continue
            if query_params.get(param) not in ['true', 'false']:
                continue
            field_name = param.replace('__isnull', '')
            if field_name not in filterset_fields:
                continue
            param_value = True if query_params.get(param) == 'true' else False
            cleaned_param_type = queryset.model._meta.get_field(field_name)
            if type(cleaned_param_type) in (ForeignKey, ManyToManyField):
                custom_filter = {param: param_value}
            else:
                continue

            if q_object is None:
                q_object = Q(**custom_filter)
            else:
                q_object &= Q(**custom_filter)
        if q_object is None:
            return queryset

        return queryset.filter(q_object)


class FilterSupportingManyToMany(
    BaseFilterBackend, CustomShemaOperationParameters
):
    def get_schema_operation_parameters(self, view):
        params = [
            {
                'name': f'{field_name}__in',
                'required': False,
                'description': 'List of uids separated by comma',
                'in': 'query',
                'schema': {'type': 'string'},
            }
            for field_name in getattr(view, 'filterset_fields', ())
            if type(view.model_class._meta.get_field(field_name))
            == ManyToManyField
        ]
        return self.return_if_not_details(view=view, value=params)

    def filter_queryset(self, request, queryset, view):
        query_params = request.query_params
        filterset_fields = getattr(view, 'filterset_fields', ())
        q_object = None
        for param in query_params:
            if not param.endswith('__in'):
                continue
            field_name = param.replace('__in', '')
            if field_name not in filterset_fields:
                continue
            cleaned_param_type = queryset.model._meta.get_field(field_name)
            if not type(cleaned_param_type) == ManyToManyField:
                continue
            values = set(query_params.get(param).split(','))

            #:  "value" must be a valid UUID4, otherwise raise ValidationError
            for value in values:
                try:
                    #:  raises ValueError if not UUID4
                    uuid.UUID(value, version=4)
                except ValueError:
                    raise ValidationError(
                        {
                            "message": f'{param}: « {value} » is not a valid UUID'
                        }
                    )

            custom_filter = {param: values}
            if q_object is None:
                q_object = Q(**custom_filter)
            else:
                q_object &= Q(**custom_filter)
        if q_object is None:
            return queryset

        return queryset.filter(q_object).distinct()



# coding: utf-8
from typing import Iterable, Dict
from django.utils import timezone
from django.http import StreamingHttpResponse


def csv_data_generator(iterable: Iterable[Dict], fields: Iterable[str]):
    """
    Generator producing UTF-8 - quoted and semicolon separated CSV
    """

    # Yields headers
    yield '{}\n'.format(
        ';'.join('"{}"'.format(field) for field in fields)
    ).encode('utf-8')

    # Yields rows
    for item in iterable:
        value = '{}\n'.format(
            ';'.join(['"{}"'.format(item.get(field, '')) for field in fields])
        ).encode('utf-8')
        yield value


def csv_streaming_response(
    request, queryset, fields: Iterable[str], filename: str = None
):
    if filename is None:
        now = timezone.now()
        filename = 'export_{}_{}.csv'.format(
            queryset.model.__name__, now.strftime("%Y-%m-%d_%H-%M")
        )

    response = StreamingHttpResponse(
        csv_data_generator(queryset, fields), content_type="text/csv"
    )
    response['Content-Disposition'] = 'attachment; filename="{}"'.format(
        filename
    )

    return response



# coding: utf-8
import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend


UserModel = get_user_model()
logger = logging.getLogger('authentication')


class ConcreteBackend(ModelBackend):
    def authenticate(
        self, request=None, username=None, password=None, **kwargs
    ):
        user = super(ConcreteBackend, self).authenticate(
            request=request, username=username, password=password, **kwargs
        )

        if user is not None:
            if kwargs.get('from_api', False):
                return user
            if self.user_can_authenticate_to_backend(user):
                return user
        return None

    def user_can_authenticate_to_backend(self, user):
        minimum_backend_auth_level = getattr(
            settings, 'MINIMUM_BACKEND_AUTH_LEVEL', 'is_superuser'
        )

        return getattr(user, minimum_backend_auth_level, False) is True



# coding: utf-8
from __future__ import unicode_literals, absolute_import, print_function
import re


def remove_prefix(value):
    return value.split('.')[1]


def replace_prefix(value):
    return value.replace('.', '_')


def camel_case_to_dash_case(value):
    return re.sub(r'([a-z])([A-Z])', r'\1-\2', value).lower()



# coding: utf-8
import random
import string

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Create users'

    def add_arguments(self, parser):
        parser.add_argument('email', type=str, help='Super user email')

    def handle(self, *args, **options):
        concrete = apps.get_app_config('concrete')
        EmailModel = concrete.models['email']
        UserModel = get_user_model()
        if options['email'] is None:
            raise ValueError("'email' is needed to create a user")

        email = options['email'].lower()

        kwargs = {
            'email': email,
            'admin': True,
            'is_superuser': True,
            'is_staff': True,
        }

        user, created = UserModel.objects.get_or_create(**kwargs)

        if created:
            password = ''.join(
                random.choice(string.ascii_letters + string.digits)  # nosec
                for _ in range(24)
            )
            user.set_password(password)
            user.save()

            port = ''
            if int(settings.PORT) not in (80, 443):
                port = ':{}'.format(settings.PORT)

            admin_url = '{}://{}{}/concrete-datastore-admin/'.format(
                settings.SCHEME, settings.HOSTNAME, port
            )
            email_instance = EmailModel(
                created_by=user,
                subject='Access to Concrete Instance',
                body='''
                Welcome to Concrete <a href="{admin_url}">{hostname}</a><br>
                <br>
                You can now connect to your concrete instance with the following
                credentials :<br>

                email {email}<br>
                password {password}<br>
                <br>
                Please change your password as you connect for the first time.

            '''.format(
                    hostname=settings.HOSTNAME,
                    admin_url=admin_url,
                    email=email,
                    password=password,
                ),
                receiver=user,
            )

            email_instance.save()



# coding: utf-8
import logging
import os

from django.conf import settings
from django.dispatch import receiver
from django.db.models.signals import pre_save

from concrete_mailer.preparers import prepare_email

from concrete_datastore.concrete.models import Email  # pylint:disable=E0611


logger = logging.getLogger(__name__)


@receiver(pre_save, sender=Email)
def send_email(sender, instance, **kwargs):
    if instance.resource_status == 'to-send':

        if not instance.subject or not instance.body or not instance.receiver:
            instance.resource_status = 'send-error'
            instance.resource_message = 'Some fields are empty'
            return

        if instance.created_by is None:
            instance.resource_status = 'send-error'
            instance.resource_message = 'No sender available'
            return
        try:
            use_tls = not getattr(settings, 'DEBUG', False)
            email = prepare_email(
                context={"sender_name": settings.EMAIL_SENDER_NAME},
                css=settings.EMAIL_CSS,
                html=instance.body,
                title=instance.subject,
                sender=settings.SERVER_EMAIL,
                recipients=[instance.receiver.email],
                use_tls=use_tls,
                email_host=settings.EMAIL_HOST,
                email_port=settings.EMAIL_PORT,
                email_host_user=settings.EMAIL_HOST_USER,
                email_host_password=settings.EMAIL_HOST_PASSWORD,
            )
            email.send()

            instance.resource_status = 'sent'
            instance.resource_message = 'Mail successfuly sent'
        except Exception:
            logger.exception('Error while sending email')
            instance.resource_status = 'send-error'
            instance.resource_message = 'Unable to send email'


def build_absolute_uri(uri):
    root = '{}://{}:{}/'.format(
        settings.SCHEME, settings.HOSTNAME, settings.PORT
    )
    if uri.startswith('/') and len(uri) > 1:
        uri = uri[1:]
    return os.path.join(root, uri)



# coding: utf-8
from django.conf import settings
from django.dispatch import receiver
from django.db.models.signals import pre_delete, post_save
from django.contrib.auth import get_user_model

import concrete_datastore.concrete.models
from concrete_datastore.concrete.models import DIVIDER_MODEL
from concrete_datastore.api.v1.views import (
    remove_instances_user_tracked_fields,
)


@receiver(pre_delete)
def on_pre_delete(sender, instance, **kwargs):
    if (
        hasattr(
            concrete_datastore.concrete.models, instance.__class__.__name__
        )
        and instance.__class__.__name__
        not in settings.IGNORED_MODELS_ON_DELETE
    ):
        # pylint: disable=no-member
        concrete_datastore.concrete.models.DeletedModel.objects.create(
            model_name=instance.__class__.__name__, uid=instance.uid
        )


@receiver(post_save, sender=get_user_model())
def on_post_save(sender, instance, **kwargs):
    if instance.level == 'blocked':
        divider_manager = getattr(
            instance, '{}s'.format(DIVIDER_MODEL.lower())
        )
        user_dividers = divider_manager.values_list('uid', flat=True)
        remove_instances_user_tracked_fields(instance, user_dividers)
        divider_manager.clear()
        instance.concrete_groups.clear()



# coding: utf-8
from django.shortcuts import HttpResponse

from social_core.exceptions import SocialAuthBaseException
from social_django.middleware import SocialAuthExceptionMiddleware
from social_core.backends.gitlab import GitLabOAuth2


class SocialCustomAuthExceptionMiddleware(SocialAuthExceptionMiddleware):
    def process_exception(self, request, exception):
        if isinstance(exception, SocialAuthBaseException):
            return HttpResponse('Denied ({})'.format(exception))
        return super().process_exception(request, exception)


class NetsachGitLabOAuth2(GitLabOAuth2):
    name = 'gitlab'
    API_URL = 'https://git.netsach.info'
    AUTHORIZATION_URL = 'https://git.netsach.info/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://git.netsach.info/oauth/token'  # nosec



# coding: utf-8
import sys
import uuid
import logging
import pendulum
import warnings

from django.conf import settings
from django.utils import timezone
from django.shortcuts import get_object_or_404
from django.contrib.auth import get_user_model, authenticate
from django.core.exceptions import ObjectDoesNotExist
from django.http.response import HttpResponseNotAllowed

from rest_framework import mixins, authentication, generics, viewsets, status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated
from rest_framework.reverse import reverse
from rest_framework.status import (
    HTTP_403_FORBIDDEN,
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
    HTTP_200_OK,
    HTTP_202_ACCEPTED,
    HTTP_201_CREATED,
    HTTP_409_CONFLICT,
)

from concrete_datastore.api.v1_1.permissions import (
    ConcreteRolesPermission,
    BlockedUsersPermission,
)
from concrete_datastore.concrete.meta import list_of_meta
from concrete_datastore.api.v1.serializers import (
    UserSerializer,
    make_serializer_class,
)
from concrete_datastore.api.v1_1.serializers import (
    make_account_me_serialier,
    ConcreteRoleSerializer,
    ConcretePermissionSerializer,
    ProcessRegisterSerializer,
    LDAPAuthLoginSerializer,
    EmailDeviceSerializer,
    TwoFactorLoginSerializer,
    BlockedUserUpdateSerializer,
)
from concrete_datastore.api.v1.authentication import (
    TokenExpiryAuthentication,
    expire_temporary_tokens,
    URLTokenExpiryAuthentication,
)
from concrete_datastore.api.v1.views import (
    validate_request_permissions,
    make_api_viewset,
    unauthorized,
    URL_TIMESTAMP,
    AccountMeApiView as ApiV1AccountMeApiView,
    ApiModelViewSet as ApiV1ModelViewSet,
    PaginatedViewSet,
)
from concrete_datastore.api.v1.permissions import (
    get_permissions_classes_by_meta_model,
)
from concrete_datastore.api.v1 import DEFAULT_API_NAMESPACE
from concrete_datastore.api.v1_1 import API_NAMESPACE

from concrete_datastore.concrete.models import (
    get_fields_and_types_of_model,
    ConcreteRole,
    ConcretePermission,
    EmailDevice,
    AuthToken,
)
from concrete_datastore.concrete.automation.signals import user_logged_in


UserModel = get_user_model()

logger = logging.getLogger('concrete-datastore')
logger_api_auth = logging.getLogger('api_auth_log')


def get_client_ip(request):
    if 'HTTP_X_FORWARDED_FOR' in request.META:
        x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


class LDAPLoginApiView(generics.GenericAPIView):
    """this view is used to login the user"""

    #: Serializer
    serializer_class = LDAPAuthLoginSerializer
    api_namespace = API_NAMESPACE

    def post(self, request, *args, **kwargs):
        """
        :param request: needs fields **email** & **password**
        :return:
          return the UserSerializer data (email, url, first_name,
          last_name, level, password and token)
        """
        serializer = LDAPAuthLoginSerializer(data=request.data)
        if not serializer.is_valid():
            return Response(
                data=serializer.errors, status=HTTP_400_BAD_REQUEST
            )

        username = serializer.validated_data["username"]

        user = authenticate(
            username=username,
            password=serializer.validated_data["password"],
            from_api=True,
        )
        if user is None:
            return Response(
                data={
                    'message': 'Wrong auth credentials',
                    "_errors": ["WRONG_AUTH_CREDENTIALS"],
                },
                status=HTTP_401_UNAUTHORIZED,
            )

        user_logged_in.send(sender=get_user_model(), user=user)

        get_user_model().objects.filter(pk=user.pk).update(
            last_login=timezone.now()
        )

        serializer = UserSerializer(
            instance=user, api_namespace=self.api_namespace
        )
        return Response(data=serializer.data, status=HTTP_200_OK)


class ProcessRegisterApiView(generics.GenericAPIView):
    serializer_class = ProcessRegisterSerializer

    def post(self, request, *args, **kwargs):

        """
        1. Check if the token is not already created and that it is valid
        2. Check if the user(process) is not already created with a different token
        """
        serializer = self.serializer_class(data=request.data)
        if not serializer.is_valid():
            return Response(
                data={'message': serializer.errors},
                status=HTTP_400_BAD_REQUEST,
            )
        application = serializer.validated_data["application"]
        instance = serializer.validated_data["instance"]
        email = "{}_{}@netsach.com".format(application, instance).lower()
        logger.info(f'Process {application} {instance} is trying to register')
        token = serializer.validated_data["token"]
        auth_token = AuthToken.objects.filter(key=token, expired=False).first()
        if auth_token:
            # If the emails match, it is authorized
            if auth_token.user.email == email:
                data = {
                    "msg": "Process successfully registered",
                    "token": token,
                    "level": auth_token.user.get_level(),
                }
                return Response(data=data, status=HTTP_202_ACCEPTED)
            # If they do not match, conflict
            else:
                data = {
                    "msg": "Email address already used with a different token"
                }
                return Response(data=data, status=HTTP_409_CONFLICT)

        password = str(uuid.uuid4())
        user, created = UserModel.objects.get_or_create(email=email)
        if created:
            user.set_password(password)
            user.set_level('blocked')
            user.save()
        else:
            # If the email address already has a token, unauthorize
            existing_token = user.auth_tokens.exclude(key=token).first()
            if existing_token is not None:
                data = {
                    "msg": "Email address already used with a different token"
                }
                return Response(data=data, status=HTTP_409_CONFLICT)
        now = pendulum.now('utc')
        now_plus_100_years = now.add(years=100)
        AuthToken.objects.create(
            key=token, user=user, expiration_date=now_plus_100_years
        )

        data = {
            "msg": "Process successfully registered",
            "token": token,
            "level": user.get_level(),
        }

        return Response(status=HTTP_202_ACCEPTED, data=data)


class ApiModelViewSet(ApiV1ModelViewSet):
    @action(
        detail=False, url_path='sets{}'.format(URL_TIMESTAMP), url_name='sets'
    )
    def get_sets(self, request, timestamp_start=None, timestamp_end=None):
        if request.parser_context["view"].model_class.__name__ == "User":
            validate_request_permissions(request=request)

        queryset, timestamp_end = self._get_queryset_filtered_since_timestamp(
            timestamp_start, timestamp_end
        )

        list_filters_field = {
            field_name: set(queryset.values_list(field_name, flat=True))
            for field_name in self.filterset_fields
        }

        data = {
            'sets': list_filters_field,
            'timestamp_start': timestamp_start or 0.0,
            'timestamp_end': timestamp_end,
        }
        return Response(data)

    def get_list_filters_field(self, queryset):
        fields_and_types = get_fields_and_types_of_model(queryset.model)

        return {
            field_name: fields_and_types[field_name]
            for field_name in self.filterset_fields
        }

    def get_extra_informations(self, queryset):
        _model_class = self.model_class or queryset.model
        extra_info = {
            'model_name': _model_class.__name__,
            'model_verbose_name': _model_class._meta.verbose_name,
            'list_display': self.get_list_display(),
            'list_filter': self.get_list_filters_field(queryset),
            'total_objects_count': queryset.count(),
            'create_url': self.request.build_absolute_uri(
                reverse("{}:{}-list".format(API_NAMESPACE, self.basename))
            ),
        }
        if _model_class.__name__ == 'User':
            extra_info['create_url'] = self.request.build_absolute_uri(
                reverse("{}:register-view".format(API_NAMESPACE))
            )
        return extra_info


class TwoFactorLoginView(generics.GenericAPIView):
    """this view is used to login the user"""

    #: Serializer
    serializer_class = TwoFactorLoginSerializer
    api_namespace = API_NAMESPACE

    def post(self, request, *args, **kwargs):
        serializer = TwoFactorLoginSerializer(data=request.data)
        if not serializer.is_valid():
            return Response(
                data=serializer.errors, status=HTTP_400_BAD_REQUEST
            )

        email = serializer.validated_data["email"]

        ip = get_client_ip(request)
        now = pendulum.now('utc').format(settings.LOGGING['datefmt'])
        user = self.request.user
        base_message = f"[{now}|{ip}|{user}|AUTH] "

        try:
            user = UserModel.objects.get(email=email.lower())
        except ObjectDoesNotExist:
            log_request = (
                base_message
                + f"Connection attempt to unknown user {email} using MFA"
            )
            logger_api_auth.info(log_request)
            return Response(
                data={
                    'message': 'Wrong auth credentials',
                    "_errors": ["WRONG_AUTH_CREDENTIALS"],
                },
                status=HTTP_401_UNAUTHORIZED,
            )
        expire_temporary_tokens(user)
        temp_token = user.temporary_tokens.filter(
            key=serializer.validated_data["token"]
        ).first()
        if not temp_token:
            log_request = (
                base_message + f"Invalid code when validating MFA for {email}"
            )
            logger_api_auth.info(log_request)
            return Response(
                data={
                    'message': 'MFA temporary token invalid',
                    "_errors": ["MFA_TEMP_TOKEN_INVALID"],
                },
                status=HTTP_401_UNAUTHORIZED,
            )
        if temp_token.expired is True:
            log_request = base_message + f"MFA temporary token expired {email}"
            logger_api_auth.info(log_request)
            return Response(
                data={
                    'message': 'MFA temporary token expired',
                    "_errors": ["MFA_TEMP_TOKEN_EXPIRED"],
                },
                status=HTTP_401_UNAUTHORIZED,
            )
        email_device = user.emaildevice_set.filter(confirmed=True).first()

        verified = email_device.verify_token(
            token=serializer.validated_data["verification_code"]
        )
        if not verified:
            log_request = (
                base_message
                + f"Connection attempt to user {email} with wrong MFA code"
            )
            logger_api_auth.info(log_request)
            return Response(
                data={
                    'message': 'Wrong verification code',
                    "_errors": ["WRONG_VERIFICATION_CODE"],
                },
                status=HTTP_401_UNAUTHORIZED,
            )

        # Expire the token to use it only once
        temp_token.expired = True
        temp_token.save()
        serializer = UserSerializer(
            instance=user,
            api_namespace=self.api_namespace,
            context={'is_verified': True},
        )
        log_request = (
            base_message
            + f"Connection attempt to user {email} using MFA is successful"
        )
        logger_api_auth.info(log_request)
        return Response(data=serializer.data, status=HTTP_200_OK)


class AccountMeApiView(ApiV1AccountMeApiView):
    def get_serializer_class(self):
        return make_account_me_serialier(self.api_namespace)


class EmailDeviceAuthView(PaginatedViewSet, viewsets.ModelViewSet):
    authentication_classes = (
        authentication.BasicAuthentication,
        authentication.SessionAuthentication,
        TokenExpiryAuthentication,
        URLTokenExpiryAuthentication,
    )
    model_class = EmailDevice
    permission_classes = (IsAuthenticated,)
    serializer_class = EmailDeviceSerializer
    filterset_fields = ('confirmed', 'email', 'user')

    def get_queryset(self):
        user = self.request.user
        # Return all Email device if at least admin
        # Else, only the user email device
        at_least_admin = False if user.is_anonymous else user.is_at_least_admin
        if at_least_admin:
            return EmailDevice.objects.all()
        else:
            return EmailDevice.objects.filter(user=user)

    def get_extra_informations(self, queryset):
        _model_class = queryset.model
        return {
            'model_name': _model_class.__name__,
            'model_verbose_name': _model_class._meta.verbose_name,
            'total_objects_count': queryset.count(),
            'create_url': self.request.build_absolute_uri(
                reverse("{}:{}-list".format(API_NAMESPACE, self.basename))
            ),
        }

    def perform_create(self, serializer):

        attrs = {'created_by': self.request.user, 'user': self.request.user}
        serializer.save(**attrs)


class ConcreteRoleApiView(PaginatedViewSet, viewsets.ModelViewSet):

    model_class = ConcreteRole
    permission_classes = (IsAuthenticated, ConcreteRolesPermission)
    api_namespace = DEFAULT_API_NAMESPACE
    serializer_class = ConcreteRoleSerializer

    def get_extra_informations(self, queryset):
        _model_class = self.model_class or queryset.model
        return {
            'model_name': _model_class.__name__,
            'model_verbose_name': _model_class._meta.verbose_name,
            'list_display': ['name'],
            'list_filter': {},
            'total_objects_count': queryset.count(),
            'create_url': self.request.build_absolute_uri(
                reverse("{}:{}-list".format(API_NAMESPACE, self.basename))
            ),
        }

    def get_queryset(self):
        return ConcreteRole.objects.all()

    def update(self, request, *args, **kwargs):
        instance = self.get_object()
        if isinstance(instance, UserModel):
            request_user = request.user
            at_least_admin = request_user.is_at_least_admin
            if not at_least_admin:
                return Response(status=HTTP_403_FORBIDDEN)
        return super(ConcreteRoleApiView, self).update(
            request, *args, **kwargs
        )

    def perform_create(self, serializer):
        attrs = {'created_by': self.request.user}
        serializer.save(**attrs)


class ConcretePermissionApiView(
    PaginatedViewSet,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
    mixins.ListModelMixin,
    viewsets.GenericViewSet,
):

    model_class = ConcretePermission
    permission_classes = (IsAuthenticated, ConcreteRolesPermission)
    serializer_class = ConcretePermissionSerializer

    def get_extra_informations(self, queryset):
        _model_class = self.model_class or queryset.model
        return {
            'model_name': _model_class.__name__,
            'model_verbose_name': _model_class._meta.verbose_name,
            'list_display': ['model_name'],
            'list_filter': {},
            'total_objects_count': queryset.count(),
            'create_url': self.request.build_absolute_uri(
                reverse("{}:{}-list".format(API_NAMESPACE, self.basename))
            ),
        }

    def get_queryset(self):
        return ConcretePermission.objects.all()

    def perform_create(self, serializer):
        attrs = {'created_by': self.request.user}
        serializer.save(**attrs)

    def update(self, request, *args, **kwargs):
        instance = self.get_object()
        if isinstance(instance, UserModel):
            request_user = request.user
            at_least_admin = request_user.is_at_least_admin
            if not at_least_admin:
                return Response(status=HTTP_403_FORBIDDEN)
        return super(ConcretePermissionApiView, self).update(
            request, *args, **kwargs
        )


for meta_model in list_of_meta:
    if meta_model.get_model_name() in ["EntityDividerModel", "UndividedModel"]:
        continue

    permissions_classes = get_permissions_classes_by_meta_model(meta_model)

    viewset_name = '{}ModelViewSet'.format(meta_model.get_model_name())

    if meta_model.get_model_name() == 'User':
        blocked_user_super_cls = make_api_viewset(
            meta_model=meta_model,
            permission_classes=(BlockedUsersPermission,),
            api_model_view_set_class=ApiModelViewSet,
            api_namespace=API_NAMESPACE,
            make_serializer_class_fct=make_serializer_class,
        )

        def get_blocked_queryset(self):
            user = self.request.user
            at_least_staff = (
                False if user.is_anonymous else user.is_at_least_staff
            )

            #: Anonymous user can only see public objects
            user_filters = {'is_active': False}
            if not at_least_staff:
                user_filters.update(public=True)
            return UserModel.objects.filter(**user_filters)

        extra_attrs = {
            'post': unauthorized,
            'update': unauthorized,
            'get_queryset': get_blocked_queryset,
        }
        blocked_user_view_cls = type(
            'BlockedUsersApiViewset', (blocked_user_super_cls,), extra_attrs
        )
        setattr(
            sys.modules[__name__],
            'BlockedUsersApiViewset',
            blocked_user_view_cls,
        )

    setattr(
        sys.modules[__name__],
        viewset_name,
        make_api_viewset(
            meta_model=meta_model,
            permission_classes=permissions_classes,
            api_model_view_set_class=ApiModelViewSet,
            api_namespace=API_NAMESPACE,
            make_serializer_class_fct=make_serializer_class,
        ),
    )


def make_deprecated_viewset(viewset, correct_url):
    class DeprecatedViewSet(viewset):
        def list(self, *args, **kwargs):
            warning_message = (
                "This route will be depricated in the next version of "
                f"concrete datastore. Please use '{correct_url}' instead"
            )
            warnings.warn(warning_message, DeprecationWarning)
            try:
                return super().list(*args, **kwargs)
            except AttributeError:
                return HttpResponseNotAllowed(self.request.method)

        def get_object(self, *args, **kwargs):
            warning_message = (
                "This route will be depricated in the next version of "
                f"concrete datastore. Please use '{correct_url}' instead"
            )
            warnings.warn(warning_message, DeprecationWarning)
            try:
                return super().get_object(*args, **kwargs)
            except AttributeError:
                return HttpResponseNotAllowed(self.request.method)

        def create(self, *args, **kwargs):
            warning_message = (
                "This route will be depricated in the next version of "
                f"concrete datastore. Please use '{correct_url}' instead"
            )
            warnings.warn(warning_message, DeprecationWarning)
            try:
                return super().create(*args, **kwargs)
            except AttributeError:
                return HttpResponseNotAllowed(self.request.method)

        def destroy(self, *args, **kwargs):
            warning_message = (
                "This route will be depricated in the next version of "
                f"concrete datastore. Please use '{correct_url}' instead"
            )
            warnings.warn(warning_message, DeprecationWarning)
            try:
                return super().destroy(*args, **kwargs)
            except AttributeError:
                return HttpResponseNotAllowed(self.request.method)

    return DeprecatedViewSet


class UnBlockUsersApiViewset(generics.GenericAPIView):
    serializer_class = BlockedUserUpdateSerializer
    authentication_classes = (
        authentication.SessionAuthentication,
        TokenExpiryAuthentication,
        URLTokenExpiryAuthentication,
    )
    permission_classes = (BlockedUsersPermission,)
    model_class = UserModel

    def get_serializer_class(self):
        return self.serializer_class

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data = self.unblock_users(serializer.data['user_uids'])
        return Response(data, status=status.HTTP_200_OK)

    def unblock_users(self, user_uids):
        #:  if one or more uids is not found or is not in queryset, raise 404
        for user_uid in user_uids:
            user = get_object_or_404(UserModel, pk=user_uid)

        data = {}
        for user_uid in user_uids:
            user = UserModel.objects.get(pk=user_uid)
            if user.is_active:
                data[user_uid] = 'User is already active'
                continue
            user.set_level('simpleuser', commit=True)
            data[user_uid] = 'User successfully unblocked'
        return data



# coding: utf-8
from django.urls import re_path

from concrete_datastore.concrete.views import email_confirmation_view
from concrete_datastore.concrete.views import unsubscribe_notifications_view
from concrete_datastore.concrete.views import (
    unsubscribe_notifications_result_view,
)

app_name = 'concrete_datastore.concrete'

urlpatterns = [
    re_path(
        r'^confirm-user-email/(?P<token>[-\w]+)',
        email_confirmation_view,
        name='email_confirmation',
    ),
    re_path(
        r'^unsubscribe-notifications/(?P<token>[-\w]+)',
        unsubscribe_notifications_view,
        name='unsubscribe_notifications',
    ),
    re_path(
        r'^unsubscribe-notifications-result/(?P<token>[-\w]+)',
        unsubscribe_notifications_result_view,
        name='unsubscribe_notifications_result',
    ),
]



# coding: utf-8
from concrete_datastore.concrete.constants import EMPTY_VALUES_MAP
from django.core.exceptions import ValidationError


def validate_file(fieldfile_obj):
    if fieldfile_obj.file is None:  # skip-test-coverage
        raise ValidationError("Error : missing file.")


def get_field_validator(field):
    # a field's value is considered INVALID if this field
    # is not allowed to be blank AND the value is empty
    empty_value = EMPTY_VALUES_MAP.get(field.f_type, {}).get(
        'empty_value', None
    )
    field_type = EMPTY_VALUES_MAP.get(field.f_type, {}).get(
        'field_type', lambda x: x
    )

    def validate_field(self, value):
        value_is_empty = field_type(value) == empty_value
        invalid = value_is_empty and not field.f_args.get('blank', False)
        if invalid:
            raise ValidationError('This field may not be blank.')
        return value

    return validate_field


def is_field_required(field):
    #
    #                 default is empty | default not empty
    #                ------------------|-------------------
    # blank is True  |      FALSE      |      FALSE
    # ---------------|-----------------|-------------------
    # blank is False |      TRUE       |      FALSE
    # ---------------|-----------------|-------------------
    blank = field.f_args.get('blank', False)
    if blank is True:
        return False

    # if field has no default value, field is required
    if 'default' not in field.f_args:
        return True

    # if there is no empty possible for the field type, it cannot be missing
    if field.f_type not in EMPTY_VALUES_MAP:
        return True

    default = field.f_args['default']
    empty = EMPTY_VALUES_MAP[field.f_type]['empty_value']

    # Field is required if default is empty
    # For BACKWARD-COMPATIBILITY
    # e.g. field STATUS has a default value, cannot be empty,
    # is not required at creation...
    return default == empty



# coding: utf-8
import os
from itertools import chain
from celery import Celery

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'settings.production')
from django.conf import settings

app = Celery('celery_app', backend='rpc://')
app.config_from_object('django.conf:settings')


def get_plugin_tasks():
    tasks = [elt['task'] for elt in settings.CELERYBEAT_SCHEDULE.values()]
    tasks.extend(settings.PLUGINS_TASKS_FUNC.keys())
    return tasks


plugin_tasks_modules = [
    plugin.rsplit('.', 1)[0]  #: module name
    for plugin in get_plugin_tasks()
    if plugin.split('.')[0] in settings.INSTALLED_PLUGINS.keys()
]

app.autodiscover_tasks(
    chain(
        ['concrete_datastore.concrete.automation.tasks'], plugin_tasks_modules
    )
)



# coding: utf-8
import logging
import os
import sys
import yaml


logger = logging.getLogger('concrete_datastore')


def load_datamodel(datamodel_path='current-datamodel.meta'):

    datamodel_path = os.getenv('DATAMODEL_FILE') or datamodel_path

    try:
        with open(datamodel_path, "r", encoding='utf8') as f:
            try:
                #:  Try to load datamodel with 'yaml.safe_load'.
                #:  This method loads both yaml and json files
                meta_model_definitions = yaml.safe_load(f)
            except yaml.scanner.ScannerError:
                message = (
                    f"Unable to parse datamodel file '{datamodel_path}'.\n"
                    "Please check that you specified the right filename"
                )
                logger.error(message)
                sys.exit(1)
    except IOError:
        message = (
            "You did not define a datamodel.\n"
            "You should either specify a path to a datamodel with "
            "DATAMODEL_FILE=<file_path>, or create a datamodel file "
            f"in ./datamodel named '{os.path.basename(datamodel_path)}'."
            "\nYou will find a sample file in ./datamodel"
        )
        logger.error(message)
        sys.exit(1)
    return meta_model_definitions


def get_log_path(filename):
    log_folder = os.getenv('LOG_FOLDER', os.path.join(os.getcwd(), 'log'))
    try:
        os.makedirs(log_folder)
    except FileExistsError:
        # Directory already exists
        pass
    return os.path.join(log_folder, filename)



# coding: utf-8
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Create users'

    def add_arguments(self, parser):
        parser.add_argument('email', type=str, help='choosen email')
        parser.add_argument('--password', type=str, help='raw password')

    def handle(self, *args, **options):
        UserClass = get_user_model()
        kwargs = {
            'email': options['email'].lower(),
            'admin': True,
            'is_superuser': True,
            'is_staff': True,
        }
        instance, created = UserClass.objects.get_or_create(**kwargs)
        if options['password'] is not None:
            instance.set_password(options['password'])
            instance.save()



# coding: utf-8
import os
from datetime import timedelta
from celery.schedules import crontab
from corsheaders.defaults import default_headers

from concrete_datastore.settings.utils import get_log_path

#: Necessary imports for ldap authentication, should not raise error when imported
try:  # nosec
    import ldap  # pylint: disable=import-error
    from django_auth_ldap.config import *  # pylint: disable=import-error
except Exception:
    pass


ALLOWED_HOSTS = []
PROJECT_ROOT = os.path.dirname(os.path.realpath(__file__)) + '/'
STATIC_ROOT = os.path.join(PROJECT_ROOT, 'static/')
MEDIA_ROOT = os.path.join(PROJECT_ROOT, 'media/')
LANGUAGE_CODE = 'fr-fr'
MEDIA_URL = '/m/'
STATIC_URL = '/s/'


STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)
PASSWORD_MIN_LENGTH = 4
PASSWORD_MIN_DIGITS = 0
PASSWORD_MIN_LOWER = 0
PASSWORD_MIN_UPPER = 0
PASSWORD_MIN_SPECIAL = 0
SPECIAL_CHARACTERS = "!@#$%%^&*()_+-=[]{}|'\""


LOGIN_REDIRECT_URL = 'index-logged'
SOCIAL_AUTH_RAISE_EXCEPTIONS = False

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'concrete_datastore.concrete.password.PasswordMinLengthValidation'
    },
    {
        'NAME': 'concrete_datastore.concrete.password.PasswordMinDigitsValidation'
    },
    {
        'NAME': 'concrete_datastore.concrete.password.PasswordMinLowerValidation'
    },
    {
        'NAME': 'concrete_datastore.concrete.password.PasswordMinUpperValidation'
    },
    {
        'NAME': 'concrete_datastore.concrete.password.PasswordMinSpecialValidation'
    },
]
DATAMODEL_VERSION = '0.0.0'

ALLOW_MULTIPLE_AUTH_TOKEN_SESSION = True

AUTH_CONFIRM_EMAIL_ENABLE = False
AUTH_CONFIRM_EMAIL_DEFAULT_REDIRECT_TO = 'https://www.netsach.org'
AUTH_CONFIRM_EMAIL_MESSAGE_BODY = """
<html>
<body>
<h3>Welcome to {platform},</h3>

<p>Please confirm your access and your email address : {email} by clicking <a rel="notrack" href='{link}'>here</a>. This email will be used to log in.</p>


<h3>Bienvenue sur {platform},</h3>

<p>Merci de confirmer votre accès et votre adresse email : {email} en cliquant <a rel="notrack" href='{link}'>ici</a>. Cet email sera utilisé pour se connecter.</p>
</body>
</html>
"""  # nosec

PASSWORD_CHANGE_TOKEN_EXPIRY_HOURS = 4

SECURE_CONNECT_EXPIRY_TIME_DAYS = 2
MAX_SECURE_CONNECT_TOKENS = 10
SECURE_TOKEN_MESSAGE_BODY = """
<html>
<body>
<h3>Welcome to {platform},</h3>

<p>Please click <a rel="notrack" href='{link}'>here</a> to authenticate to the platform.</p>


<h3>Bienvenue sur {platform},</h3>

<p>Veuillez cliquer <a rel="notrack" href='{link}'>ici</a> pour vous connecter sur la plateforme.</p>
</body>
</html>
"""  # nosec

DEFAULT_RESET_PASSWORD_URL_FORMAT = (
    '/#/reset-password/{token}/{email}/'  # nosec
)

AUTH_CONFIRM_RESET_PASSWORD_EMAIL_BODY = """
<html>
<body>
<h3>Reset password</h3>
<p>Please follow <a rel="notrack" href="{link}">this link</a> to reset your password.<br>
Ignore this email if you didn't ask to reset your password.</p><br>

<h3>Mise à jour du mot de passe</h3>
<p>Veuillez suivre <a rel="notrack" href="{link}">ce lien</a> pour mettre à jour votre mot de passe<br>
Veuillez ignorer ce mail si vous n'avez pas demandé à mettre à jour votre mot de passe</p>
</body>
</html>
"""  # nosec

PLATFORM_NAME = 'Concrete Datastore'

AUTHENTICATION_BACKENDS = [
    'concrete_datastore.authentication.auth.ConcreteBackend',
    'concrete_datastore.authentication.oauth2_utils.NetsachGitLabOAuth2',
]

MINIMUM_BACKEND_AUTH_LEVEL = 'is_superuser'
MINIMUM_LEVEL_FOR_USER_LIST = 'is_at_least_staff'

ROOT_URLCONF = 'concrete_datastore.routes.urls'

PASSWORD_EXPIRY_TIME = 0  # in days, 0 for no expiration


API_TOKEN_EXPIRY = 0  # in minutes, 0 for no expiration
EXPIRY_EXTRA_PERIOD = 0  # in minutes, 0 for no extra period

CORS_ORIGIN_ALLOW_ALL = True
CORS_ALLOW_HEADERS = default_headers + (
    'Access-Control-Allow-Headers',
    'Access-Control-Allow-Origin',
    'X-CSRFToken',
    'Authorization',
    'X-Entity-Uid',
    'Content-Type',
    'Cache-Control',
)
CORS_ORIGIN_WHITELIST = "*"


MIDDLEWARE = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'social_django.middleware.SocialAuthExceptionMiddleware',
    'concrete_datastore.authentication.oauth2_utils.SocialCustomAuthExceptionMiddleware',
    'concrete_datastore.concrete.middleware.OTPCustomMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': (
            os.path.join(os.path.dirname(__file__), "templates"),
            os.path.join(
                os.path.dirname(os.path.dirname(__file__)), "doc/templates"
            ),
            os.path.join(
                os.path.dirname(os.path.dirname(__file__)),
                "concrete/templates",
            ),
        ),
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.request',
                'social_django.context_processors.backends',
                'social_django.context_processors.login_redirect',
            ]
        },
    }
]


TIME_ZONE = 'UTC'
LANGUAGE_CODE = 'en-en'

USE_I18N = True

USE_L10N = True

USE_TZ = True

API_MAX_PAGINATION_SIZE = 250
API_MAX_PAGINATION_SIZE_NESTED = 125
DEFAULT_PAGE_SIZE = 250

# DRF
REST_FRAMEWORK = {
    'DEFAULT_SCHEMA_CLASS': 'concrete_datastore.interfaces.openapi_schema_generator.AutoSchema',
    'DATETIME_FORMAT': "%Y-%m-%dT%H:%M:%SZ",
    'COERCE_DECIMAL_TO_STRING': False,
    'PAGE_SIZE': DEFAULT_PAGE_SIZE,
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.SessionAuthentication',
        'concrete_datastore.api.v1.authentication.TokenExpiryAuthentication',
        'concrete_datastore.api.v1.authentication.URLTokenExpiryAuthentication',
    ),
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.LimitOffsetPagination',
    'DEFAULT_RENDERER_CLASSES': (
        'rest_framework.renderers.JSONRenderer',
        'rest_framework.renderers.BrowsableAPIRenderer',
    ),
}

AUTH_USER_MODEL = 'concrete.User'

API_REGISTER_EMAIL_FILTER = r'.*'

ADMIN_SHOW_USER_PERMISSIONS = False

ADMIN_HEADER = "Concrete Backoffice"

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.sites',
    'django_extensions',
    'social_django',
    'django_otp',
    'rest_framework',
    'rest_framework_gis',
    'django.contrib.gis',
    'corsheaders',
    'concrete_datastore.concrete',
]

LICENSE = 'GNU GENERAL PUBLIC LICENSE Version 3'

OPENAPI_SPEC_TITLE = os.environ.get('APP_INSTANCE', 'Concrete Datastore')

DEFAULT_LOGGING_FORMAT = (
    "%(asctime)s § %(levelname)5s [%(name)s] : %(message)s "
    "([%(filename)s:%(lineno)s %(funcName)s)"
)
DEFAULT_LOGGING_LEVEL = "INFO"

LOGGING_FORMAT = os.environ.get('LOGGING_FORMAT', DEFAULT_LOGGING_FORMAT)
LOGGING_LEVEL = os.environ.get('LOGGING_LEVEL', DEFAULT_LOGGING_LEVEL).upper()

LOGGING = {
    'version': 1,
    'datefmt': "YY-MM-DD HH:mm:ss",
    'formatters': {'verbose': {'format': DEFAULT_LOGGING_FORMAT}},
    'filters': {
        'require_debug_true': {'()': 'django.utils.log.RequireDebugTrue'}
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'filters': ['require_debug_true'],
            'class': 'logging.StreamHandler',
            'formatter': 'verbose',
        },
        'stream': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'verbose',
        },
        'safe_file': {
            'level': 'INFO',
            'class': 'logging.handlers.WatchedFileHandler',
            'filename': get_log_path('read.log'),
        },
        'unsafe_file': {
            'level': 'INFO',
            'class': 'logging.handlers.WatchedFileHandler',
            'filename': get_log_path('action.log'),
        },
        'auth_file': {
            'level': 'INFO',
            'class': 'logging.handlers.WatchedFileHandler',
            'filename': get_log_path('auth.log'),
        },
    },
    'loggers': {
        '': {
            'handlers': ['console', 'stream'],
            'propagate': True,
            'level': 'INFO',
        },
        'concrete_datastore': {
            'handlers': ['console', 'stream'],
            'propagate': False,
            'level': 'DEBUG',
        },
        'django.db.backends': {
            'handlers': ['console', 'stream'],
            'propagate': False,
            'level': 'WARNING',
        },
        'django.request': {
            'handlers': ['console', 'stream'],
            'propagate': False,
            'level': 'WARNING',
        },
        'celery.task': {
            'handlers': ['console', 'stream'],
            'propagate': False,
            'level': 'WARNING',
        },
        'celery.app.trace': {
            'handlers': ['console', 'stream'],
            'propagate': False,
            'level': 'INFO',
        },
        'api_safe_log': {
            'handlers': ['safe_file'],
            'propagate': True,
            'level': 'INFO',
        },
        'api_unsafe_log': {
            'handlers': ['unsafe_file'],
            'propagate': True,
            'level': 'INFO',
        },
        'api_auth_log': {
            'handlers': ['auth_file'],
            'propagate': True,
            'level': 'INFO',
        },
    },
}
# Dict:
# Key: Task path; Value: Queue name
PLUGINS_TASKS_FUNC = {
    # 'concrete_datastore.dummy_plugin.automation.tasks.dummy_task': 'default_queue',
}


EMAIL_CSS = ''
EMAIL_SENDER_NAME = '<Concrete Datastore>'
EMAIL_HOST = ''
# Limit of different values before deactivating filter
LIMIT_DEACTIVATE_FILTER_IN_ADMIN = 50

BROKER_URL = os.environ.get('BROKER_URL', 'redis://localhost:6379/0')
PLUGIN_TASK_TIMEDELTA_SEC = int(
    os.environ.get('PLUGIN_TASK_TIMEDELTA_SEC', 20)
)
BROKER_TRANSPORT_OPTIONS = {
    'fanout_prefix': True,
    'fanout_patterns': True,
    'visibility_timeout': 24 * 60 * 60,
}

INSTALLED_PLUGINS = {
    # 'plugin_concrete_xxx': '1.0.0'
}

CELERYBEAT_SCHEDULE = {
    'async_run_plugin_tasks': {
        'task': 'concrete_datastore.concrete.automation.tasks.async_run_plugin_tasks',
        'schedule': timedelta(seconds=PLUGIN_TASK_TIMEDELTA_SEC),
        'options': {'queue': 'periodic'},
    }
}

USE_CONCRETE_ROLES = False
USE_CORE_AUTOMATION = False
# Example:
# CONCRETE_SCOPES_FILTER_LOOKUP_FOR_UNSUBSCRIBE_JSON = '{"archived": false}'
CONCRETE_SCOPES_FILTER_LOOKUP_FOR_UNSUBSCRIBE_JSON = '{}'
# Populate the Django user from the LDAP directory.
AUTH_LDAP_USER_ATTR_MAP = {
    'email': 'mail',
    "first_name": "givenName",
    "last_name": "sn",
}
AUTH_LDAP_USER_QUERY_FIELD = 'email'

AUTH_LDAP_BIND_AS_AUTHENTICATING_USER = True

# This is the default, but I like to be explicit.
AUTH_LDAP_ALWAYS_UPDATE_USER = True

# Use LDAP group membership to calculate group permissions.
AUTH_LDAP_FIND_GROUP_PERMS = False

# Cache distinguished names and group memberships for an hour to minimize
# LDAP traffic.
AUTH_LDAP_CACHE_TIMEOUT = 3600
USE_AUTH_LDAP = False


SOCIAL_AUTH_GITLAB_KEY = ''
SOCIAL_AUTH_GITLAB_SECRET = ''  # nosec

USE_MULTIPLE_TOKENS = False
# If true, return a new token at each login

SESSIONS_NUMBER_CONTROL_ENABLED = False
# Is used only if USE_MULTIPLE_TOKENS == True

MAX_SIMULTANEOUS_SESSIONS = 1
# 0 for unlimited, only if SESSIONS_NUMBER_CONTROL_ENABLED == True

TWO_FACTOR_CODE_TIMEOUT_SECONDS = 600
TWO_FACTOR_TOKEN_MSG = """
<html>
<body>
<h3>Verification code - {platform_name}</h3>
<p>Please enter the following confirmation code to authenticate to the platform: <br>
<strong>{confirm_code}</strong><br>
This code is valid for {min_validity} minutes.
</p>

<h3>Code de vérification - {platform_name}</h3>
<p>Veuillez entrer le code de confirmation suivant pour vous connecter sur la plateforme: <br>
<strong>{confirm_code}</strong><br>
Ce code est valable pendant {min_validity} minutes.
</p>
</body>
</html>
"""  # nosec
USE_TWO_FACTOR_AUTH = False
MFA_RULE_PER_USER = 'concrete_datastore.api.v1.authentication.default_mfa_rule'

SESSION_CONTROL_DEFAULT_RULE = (
    'concrete_datastore.api.v1.serializers.default_session_control_exempt_rule'
)

# Remote concrete authentication
REMOTE_CONCRETE_AUTH_ENABLED = False
URL_REMOTE_DATASTORE = 'http://xxx:8000/api/v1.1/'
AUTH_REMOTE_URI = 'auth/login/'
USERNAME_AUTH_KEY = 'email'

# Backend login group creation rule
BACKEND_GROUP_CREATION_RULE = 'concrete_datastore.api.v1.authentication.default_backend_group_creation_rule'

ALLOW_SEND_EMAIL_ON_REGISTER = True

REGISTER_EMAIL_SUBJECT = "Account created"

DEFAULT_REGISTER_URL_FORMAT = '/#/set-password/{token}/{email}/'  # nosec

DEFAULT_REGISTER_EMAIL_FORMAT = """
<html>
<body>
<h3>Set your password</h3>
<p>Please follow <a rel="notrack" href="{link}">this link</a> to set your password and complete your register process.<p><br>

<h3>Sélectionner votre mot de passe</h3>
<p>Veuillez suivre <a rel="notrack" href="{link}">ce lien</a> pour choisir votre mot de passe et compléter votre inscription.</p><br>
</body>
</html>
"""  # nosec

# Flag to allow a user to reuse a password on change, only applicable if
# current password is not expired
ALLOW_REUSE_PASSWORD_ON_CHANGE = False

IGNORED_MODELS_ON_DELETE = [
    "DeletedModel",
    "AuthToken",
    "TemporaryToken",
    "SecureConnectToken",
]
ADMIN_URL_ENABLED = True
ADMIN_ROOT_URI = "concrete-datastore-admin"



# coding: utf-8
from importlib import import_module

from django.conf import settings
from django.http import JsonResponse, HttpResponse

from rest_framework.views import APIView
from rest_framework.renderers import OpenAPIRenderer

import concrete_datastore
from concrete_datastore.interfaces.openapi_schema_generator import (
    SchemaGenerator,
)


def service_status_view(request):
    plugins = {}
    list_plugins = []
    # Get a unique module
    if isinstance(settings.PLUGINS_TASKS_FUNC, dict):
        list_plugins = settings.PLUGINS_TASKS_FUNC.keys()

        for path_task in list_plugins:
            task_split = path_task.rsplit('.')
            module = import_module(task_split[0])
            plugins[module.__name__] = module.__version__

    return JsonResponse(
        {
            'version': concrete_datastore.__version__,
            'datamodel_version': settings.DATAMODEL_VERSION,
            'api': concrete_datastore.api.v1_1.__version__,
            'plugins': plugins,
            'healthy': True,
            'message': '',
            'name': 'concrete-datastore',
            'license': 'All rights reserved. Netsach 2019.',
        }
    )


class OpenApiView(APIView):
    patterns = None

    def get(self, request, spec_format):
        url_scheme = request._request.META['wsgi.url_scheme']
        try:
            http_host = request._request.META['HTTP_HOST']
        except KeyError:
            http_host = (
                f"{request._request.META['REMOTE_ADDR']}"
                f":{request._request.META['SERVER_PORT']}"
            )

        user_auth_header = request._request.headers.get('Authorization')
        if user_auth_header:
            user_token = user_auth_header.replace('Token ', '')
            from_db = True
        else:
            user_token = None
            from_db = False
        server = f'{url_scheme}://{http_host}'

        generator = SchemaGenerator(
            servers=[server],
            patterns=self.patterns,
            user_token=user_token,
            from_db=from_db,
        )
        schema = generator.get_schema(request=request)

        if spec_format == 'yaml':
            spec = OpenAPIRenderer().render(schema, renderer_context={})
            return HttpResponse(spec.decode())

        return JsonResponse(schema)



# coding: utf-8
from django.conf import settings

from concrete_datastore.api.v1.exceptions import (
    PasswordInsecureValidationError,
)


class PasswordMinLengthValidation(object):
    """
    Validate whether the password is of a minimum length.
    """

    code = 'NOT_ENOUGH_CHARS'

    def __init__(self):
        self.min_length = settings.PASSWORD_MIN_LENGTH

    def validate(self, password, user=None):
        if len(password) < self.min_length:
            raise PasswordInsecureValidationError(
                message="The password must contain at least {} character(s).".format(
                    self.min_length
                ),
                code=self.code,
                params={'min_length': self.min_length},
            )

    def get_help_text(self):
        help_text = (
            "Your password must contain at least {min_length} character(s)."
        ).format(min_length=self.min_length)
        return help_text

    def get_help_text_fr(self):
        help_text = (
            "Votre mot de passe doit contenir au moins {min_length} caractère(s)"
        ).format(min_length=self.min_length)
        return help_text


class PasswordMinDigitsValidation(object):
    """
    Validate whether the password is of a minimum length.
    """

    code = 'NOT_ENOUGH_DIGITS'

    def __init__(self):
        self.min_digits = settings.PASSWORD_MIN_DIGITS

    def get_digits_count(self, password):
        count = 0
        for char in password:
            if char.isdigit():
                count += 1
        return count

    def validate(self, password, user=None):
        if self.get_digits_count(password) < self.min_digits:
            raise PasswordInsecureValidationError(
                message="The password must contain at least {} digit(s).".format(
                    self.min_digits
                ),
                code=self.code,
                params={'min_digits': self.min_digits},
            )

    def get_help_text(self):
        help_text = (
            "Your password must contain at least {min_digits} digit(s)."
        ).format(min_digits=self.min_digits)
        return help_text

    def get_help_text_fr(self):
        help_text = (
            "Votre mot de passe doit contenir au moins {min_digits} chiffre(s)"
        ).format(min_digits=self.min_digits)
        return help_text


class PasswordMinLowerValidation(object):
    """
    Validate whether the password is of a minimum length.
    """

    code = 'NOT_ENOUGH_LOWER'

    def __init__(self):
        self.min_lower = settings.PASSWORD_MIN_LOWER

    def get_lower_count(self, password):
        count = 0
        for char in password:
            if char.isdigit() is False and char.lower() == char:
                count += 1
        return count

    def validate(self, password, user=None):
        if self.get_lower_count(password) < self.min_lower:
            raise PasswordInsecureValidationError(
                message="The password must contain at least {} lower character(s).".format(
                    self.min_lower
                ),
                code=self.code,
                params={'min_lower': self.min_lower},
            )

    def get_help_text(self):
        help_text = (
            "Your password must contain at least {min_lower} lower character(s)."
        ).format(min_lower=self.min_lower)
        return help_text

    def get_help_text_fr(self):
        help_text = (
            "Votre mot de passe doit contenir au moins {min_lower} caractère(s) minuscule(s)"
        ).format(min_lower=self.min_lower)
        return help_text


class PasswordMinUpperValidation(object):
    """
    Validate whether the password is of a minimum length.
    """

    code = 'NOT_ENOUGH_UPPER'

    def __init__(self):
        self.min_upper = settings.PASSWORD_MIN_UPPER

    def get_upper_count(self, password):
        count = 0
        for char in password:
            if char.isdigit() is False and char.upper() == char:
                count += 1
        return count

    def validate(self, password, user=None):
        if self.get_upper_count(password) < self.min_upper:
            raise PasswordInsecureValidationError(
                message="The password must contain at least {} upper character(s).".format(
                    self.min_upper
                ),
                code=self.code,
                params={'min_upper': self.min_upper},
            )

    def get_help_text(self):
        help_text = (
            "Your password must contain at least {min_upper} upper character(s)."
        ).format(min_upper=self.min_upper)
        return help_text

    def get_help_text_fr(self):
        help_text = (
            "Votre mot de passe doit contenir au moins {min_upper} caractère(s) majuscule(s)"
        ).format(min_upper=self.min_upper)
        return help_text


class PasswordMinSpecialValidation(object):
    """
    Validate whether the password has a minimum of special characters.
    """

    code = 'NOT_ENOUGH_SPECIAL'

    def __init__(self):
        self.min_special = settings.PASSWORD_MIN_SPECIAL

    def get_special_count(self, password):
        count = 0
        for char in password:
            if char in settings.SPECIAL_CHARACTERS:
                count += 1
        return count

    def validate(self, password, user=None):
        if self.get_special_count(password) < self.min_special:
            raise PasswordInsecureValidationError(
                message=(
                    "The password must contain at least {min_special} special"
                    " character(s) from these : {special_list}"
                ).format(
                    min_special=self.min_special,
                    special_list=''.join(set(settings.SPECIAL_CHARACTERS)),
                ),
                code=self.code,
                params={'min_special': self.min_special},
            )

    def get_help_text(self):
        help_text = (
            "Your password must contain at least {min_special} special"
            " character(s) from these : {special_list}"
        ).format(
            min_special=self.min_special,
            special_list=''.join(set(settings.SPECIAL_CHARACTERS)),
        )
        return help_text

    def get_help_text_fr(self):
        help_text = (
            "Votre mot de passe doit contenir au moins {min_special} caractères spéciaux parmi ceux là: {special_list}"
        ).format(
            min_special=self.min_special,
            special_list=''.join(set(settings.SPECIAL_CHARACTERS)),
        )
        return help_text



# coding: utf-8
from django.urls import re_path
from django.conf import settings

from rest_framework.routers import SimpleRouter

from concrete_datastore.concrete.meta import list_of_meta
from concrete_datastore.api.v1.views import (
    LoginApiView,
    RegisterApiView,
    ResetPasswordApiView,
    ChangePasswordView,
    RetrieveSecureTokenApiView,
    SecureLoginApiView,
    GenerateSecureTokenApiView,
)
from concrete_datastore.api.v1_1.views import (  # pylint:disable=E0611
    ConcreteRoleApiView,
    ConcretePermissionApiView,
    EmailDeviceAuthView,
    LDAPLoginApiView,
    TwoFactorLoginView,
    UnBlockUsersApiViewset,
    BlockedUsersApiViewset,
    AccountMeApiView,
    ProcessRegisterApiView,
)
from concrete_datastore.api.v1_1 import views, API_NAMESPACE

app_name = 'concrete_datastore.concrete'

# API Front end
router = SimpleRouter()

for meta_model in list_of_meta:
    if meta_model.get_model_name() in ["EntityDividerModel", "UndividedModel"]:
        continue

    viewset = getattr(
        views, '{}ModelViewSet'.format(meta_model.get_model_name())
    )

    router.register(
        prefix=meta_model.get_dashed_case_class_name(),
        viewset=viewset,
        basename=meta_model.get_dashed_case_class_name(),
    )


router.register(
    prefix='acl/role', viewset=ConcreteRoleApiView, basename='acl-role'
)
router.register(
    prefix='acl/permission',
    viewset=ConcretePermissionApiView,
    basename='acl-permission',
)
router.register(
    prefix='blocked-users',
    viewset=BlockedUsersApiViewset,
    basename='blocked-users',
)

router.register(
    prefix='email-device', viewset=EmailDeviceAuthView, basename='email-device'
)

specific_urlpatterns = [
    re_path(
        r'^auth/ldap/login', LDAPLoginApiView.as_view(), name='login-ldap-view'
    ),
    re_path(
        r'auth/two-factor/login',
        TwoFactorLoginView.as_view(),
        name='two-factor-login',
    ),
    re_path(
        r'^auth/login/',
        LoginApiView.as_view(api_namespace=API_NAMESPACE),
        name='login-view',
    ),
    re_path(
        r'^auth/register/',
        RegisterApiView.as_view(api_namespace=API_NAMESPACE),
        name='register-view',
    ),
    re_path(
        r'^account/me/',
        AccountMeApiView.as_view(api_namespace=API_NAMESPACE),
        name='account-me',
    ),
    re_path(
        r'^auth/change-password/',
        ChangePasswordView.as_view(api_namespace=API_NAMESPACE),
        name='change-password',
    ),
    re_path(
        r'^auth/reset-password/',
        ResetPasswordApiView.as_view(),
        name='reset-password',
    ),
    re_path(
        r'secure-connect/retrieve-token',
        RetrieveSecureTokenApiView.as_view(),
        name='retrieve-secure-token',
    ),
    re_path(
        r'secure-connect/login/',
        SecureLoginApiView.as_view(api_namespace=API_NAMESPACE),
        name='secure-connect-login',
    ),
    re_path(
        r'secure-connect/generate-token',
        GenerateSecureTokenApiView.as_view(),
        name='generate-secure-token',
    ),
    re_path(
        r'process/register/',
        ProcessRegisterApiView.as_view(),
        name='register-as-process',
    ),
    re_path(
        r'unblock-users',
        UnBlockUsersApiViewset.as_view(),
        name='unblock-users',
    ),
]
if settings.USE_AUTH_LDAP:
    specific_urlpatterns += [
        re_path(
            r'^auth/ldap/login',
            LDAPLoginApiView.as_view(),
            name='login-ldap-view',
        )
    ]


urlpatterns = router.urls + specific_urlpatterns



# coding: utf-8
from __future__ import unicode_literals, absolute_import, print_function

AUTHORIZED_IP_PROTOCOLS = ("ipv4", "ipv6", "ipv4_6")
PROTOCOL_EQUIVALENCE = {"ipv4": "IPv4", "ipv6": "IPv6", "ipv4_6": "both"}

STD_MODEL_SPECIFIER_KEYS = (
    "std.specifier",
    "std.name",
    "std.verbose_name",
    # "std.verbose_name_plural",
    "std.description",
    "std.fields",
)

CONCRETE_COMMON_FIELDS = (
    'uid',
    'modification_date',
    'creation_date',
    'public',
)

CONCRETE_USER_PROTECTED_FIELDS = (
    'superuser',
    'level',
    'admin',
    'is_at_least_admin',
    'is_at_least_staff',
    'password_modification_date',
    'password_modification_token',
    'subscription_notification_token',
    'login_counter',
    'unsubscribe_to',
    'unsubscribe_all',
) + CONCRETE_COMMON_FIELDS

CONCRETE_MODELS_PROTECTED_FIELDS = (
    'created_by',
    'can_admin_users',
    'can_view_users',
    'can_admin_groups',
    'can_view_groups',
) + CONCRETE_COMMON_FIELDS

CONCRETE_CUSTOM_MODELS = (
    'TemporaryToken',
    'ConcreteRole',
    'ConcretePermission',
    'SecureConnectToken',
    'UserConfirmation',
    'ConfirmableUserAbstract',
    'HasPermissionAbstractUser',
    'PasswordChangeToken',
    'DefaultDivider',
    'DeletedModel',
)

STD_FIELD_SPECIFIER_KEYS = (
    "std.specifier",
    "std.name",
    "std.verbose_name",
    "std.description",
    "std.type",
)

STD_SPECIFIER = {
    "Model": STD_MODEL_SPECIFIER_KEYS,
    "Field": STD_FIELD_SPECIFIER_KEYS,
}

STD_MODEL_SPECIFIER_KEYS_V1 = ("fields", "name", "uid")
STD_FIELD_SPECIFIER_KEYS_V1 = ("datatype", "name", "attributes")

STD_SPECIFIER_V1 = {
    "Model": STD_MODEL_SPECIFIER_KEYS_V1,
    "Field": STD_FIELD_SPECIFIER_KEYS_V1,
}

TYPE_EQ_V1 = {
    'bool': 'BooleanField',
    'txt': 'TextField',
    'char': 'CharField',
    'url': 'URLField',
    'int': 'IntegerField',
    'bigint': 'BigIntegerField',
    'uid': 'UUIDField',
    'float': 'FloatField',
    'decimal': 'DecimalField',
    'json': 'JSONField',
    'file': 'FileField',
    'image': 'ImageField',
    'datetime': 'DateTimeField',
    'date': 'DateField',
    'email': 'EmailField',
    'fk': 'ForeignKey',
    'm2m': 'ManyToManyField',
    'ip': 'GenericIPAddressField',
    'point': 'PointField',
}


#:  Generic class attributes.
#:  When adding a new version, Class attributes should be added here
VERSIONS_ATTRIBUTES = {
    '1.0.0': {
        'field_type_spec': 'datatype',
        'fields_spec': 'fields',
        'element_name': 'name',
        'element_id': 'uid',
        'equivalence_table': TYPE_EQ_V1,
        'std_verifier': STD_SPECIFIER_V1,
    }
}



# coding: utf-8
from django.core.exceptions import ValidationError


class ConcreteBaseError(ValueError):
    pass


class PasswordInsecureValidationError(ValidationError, ConcreteBaseError):
    detail = 'PASSWORD_INSECURE'


class WrongEntityUIDError(ConcreteBaseError):
    detail = 'WRONG_ENTITY_UID'



# coding: utf-8
from django.apps import apps
from django.core.management.base import BaseCommand
from django.contrib.auth import get_user_model

from concrete_datastore.api.v1.views import (
    remove_instances_user_tracked_fields,
)
from concrete_datastore.concrete.models import (
    DIVIDER_MODEL,
    ConcretePermission,
)
from concrete_datastore.concrete.meta import meta_models


class Command(BaseCommand):
    help = 'Init Applications (create groups, permissions, etc.)'

    def handle(self, *args, **options):
        UserModel = get_user_model()
        for user_blocked in UserModel.objects.filter(is_active=False):
            divider_manager = getattr(
                user_blocked, '{}s'.format(DIVIDER_MODEL.lower())
            )
            user_dividers = divider_manager.values_list('uid', flat=True)
            remove_instances_user_tracked_fields(user_blocked, user_dividers)
            divider_manager.clear()
            user_blocked.concrete_groups.clear()

        all_dividers = apps.get_model(
            "concrete.{}".format(DIVIDER_MODEL)
        ).objects.all()
        for user_active in UserModel.objects.filter(is_active=True):
            divider_manager = getattr(
                user_active, '{}s'.format(DIVIDER_MODEL.lower())
            )
            user_dividers = divider_manager.all()
            dividers_to_remove = all_dividers.difference(user_dividers)
            remove_instances_user_tracked_fields(
                user_active, dividers_to_remove.values_list('uid', flat=True)
            )

        for meta_model in meta_models:
            if meta_model.get_model_name() in [
                "EntityDividerModel",
                "UndividedModel",
            ]:
                continue
            model_name = meta_model.get_model_name()
            # pylint: disable=no-member
            ConcretePermission.objects.get_or_create(model_name=model_name)



# coding: utf-8
from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model

from rest_framework.reverse import reverse
from rest_framework import serializers

from concrete_datastore.concrete.models import (
    DIVIDER_MODEL,
    ConcreteRole,
    ConcretePermission,
    EmailDevice,
)
from concrete_datastore.api.v1 import DEFAULT_API_NAMESPACE
from concrete_datastore.api.v1.serializers import make_serializer_class
from concrete_datastore.api.v1.signals import build_absolute_uri
from concrete_datastore.concrete.meta import get_meta_definition_by_model_name

concrete = apps.get_app_config('concrete')


class ProcessRegisterSerializer(serializers.Serializer):
    application = serializers.CharField(max_length=200, required=True)
    instance = serializers.CharField(max_length=200, required=True)
    token = serializers.UUIDField(required=True)


def make_account_me_serialier(api_namespace=DEFAULT_API_NAMESPACE):
    meta = get_meta_definition_by_model_name('User')
    meta_serializer_class = make_serializer_class(
        meta_model=meta, api_namespace=api_namespace, nested=False, safe=True
    )
    Meta = meta_serializer_class.Meta
    Meta.fields = [
        field
        for field in Meta.fields
        if field not in ['admin', 'is_staff', 'is_superuser', 'is_active']
    ] + ['level']

    Meta.read_only_fields = Meta.read_only_fields + [
        'uid',
        'public',
        'creation_date',
        'modification_date',
        '{}s'.format(DIVIDER_MODEL.lower()),
    ]

    serializer_attrs = {
        'Meta': Meta,
        'level': serializers.SerializerMethodField(),
        'get_level': lambda self, obj: obj.get_level(),
    }

    if settings.USE_CONCRETE_ROLES:
        Meta.fields = Meta.fields + ['roles', 'roles_uid']
        Meta.read_only_fields = Meta.read_only_fields + ['roles', 'roles_uid']
        serializer_attrs.update(
            {
                'roles': serializers.SerializerMethodField(),
                'get_roles': lambda self, obj: obj.get_roles(),
                'roles_uid': serializers.SerializerMethodField(),
                'get_roles_uid': lambda self, obj: obj.get_roles_uid(),
            }
        )
    return type(
        str('AccountMeSerializer'), (meta_serializer_class,), serializer_attrs
    )


class ConcreteRoleSerializer(serializers.ModelSerializer):
    UserModel = get_user_model()
    users_uid = serializers.PrimaryKeyRelatedField(
        source="users",
        many=True,
        required=False,
        queryset=UserModel.objects.all(),
    )
    url = serializers.SerializerMethodField()

    class Meta:
        model = ConcreteRole
        fields = (
            "uid",
            "name",
            "users_uid",
            "url",
            "created_by",
            "creation_date",
            "modification_date",
        )
        read_only_fields = (
            'uid',
            'users',
            'url',
            'created_by',
            'creation_date',
            'modification_date',
        )

    def get_url(self, obj):
        uri = reverse("api_v1_1:acl-role-detail", args=(obj.pk,))
        if hasattr(self, '_context'):
            if 'request' in self._context:
                request = self._context['request']
                return request.build_absolute_uri(uri)
        return build_absolute_uri(uri)  # skip-test-coverage


class EmailDeviceSerializer(serializers.ModelSerializer):

    url = serializers.SerializerMethodField()

    class Meta:
        model = EmailDevice
        fields = (
            "uid",
            "name",
            "email",
            "confirmed",
            "url",
            'user',
            "created_by",
            "creation_date",
            "modification_date",
        )
        read_only_fields = (
            'uid',
            'url',
            'user',
            'created_by',
            'creation_date',
            'modification_date',
        )

    def get_url(self, obj):
        uri = reverse("api_v1_1:email-device-detail", args=(obj.pk,))
        if hasattr(self, '_context'):
            if 'request' in self._context:
                request = self._context['request']
                return request.build_absolute_uri(uri)
        return build_absolute_uri(uri)  # skip-test-coverage


class ConcretePermissionSerializer(serializers.ModelSerializer):

    create_roles = serializers.StringRelatedField(many=True, read_only=True)
    update_roles = serializers.StringRelatedField(many=True, read_only=True)
    retrieve_roles = serializers.StringRelatedField(many=True, read_only=True)
    delete_roles = serializers.StringRelatedField(many=True, read_only=True)
    url = serializers.SerializerMethodField()

    def get_url(self, obj):
        uri = reverse("api_v1_1:acl-permission-detail", args=(obj.pk,))
        if hasattr(self, '_context'):
            if 'request' in self._context:
                request = self._context['request']
                return request.build_absolute_uri(uri)
        return build_absolute_uri(uri)  # skip-test-coverage

    create_roles_uid = serializers.PrimaryKeyRelatedField(
        source="create_roles",
        many=True,
        required=False,
        queryset=ConcreteRole.objects.all(),
    )
    update_roles_uid = serializers.PrimaryKeyRelatedField(
        source="update_roles",
        many=True,
        required=False,
        queryset=ConcreteRole.objects.all(),
    )
    retrieve_roles_uid = serializers.PrimaryKeyRelatedField(
        source="retrieve_roles",
        many=True,
        required=False,
        queryset=ConcreteRole.objects.all(),
    )
    delete_roles_uid = serializers.PrimaryKeyRelatedField(
        source="delete_roles",
        many=True,
        required=False,
        queryset=ConcreteRole.objects.all(),
    )

    class Meta:
        model = ConcretePermission
        fields = (
            "uid",
            "model_name",
            "create_roles",
            "update_roles",
            "retrieve_roles",
            "delete_roles",
            "create_roles_uid",
            "update_roles_uid",
            "retrieve_roles_uid",
            "delete_roles_uid",
            "url",
        )
        read_only_fields = (
            'uid',
            'users',
            "create_roles",
            "update_roles",
            "retrieve_roles",
            "delete_roles",
            'url',
            'created_by',
            'creation_date',
            'modification_date',
        )


class LDAPAuthLoginSerializer(serializers.Serializer):
    username = serializers.CharField()
    password = serializers.CharField()


class TwoFactorLoginSerializer(serializers.Serializer):
    email = serializers.EmailField()
    token = serializers.CharField()
    verification_code = serializers.CharField()


class BlockedUserUpdateSerializer(serializers.Serializer):
    user_uids = serializers.ListField(child=serializers.UUIDField())



# coding: utf-8
from importlib import import_module

from django.db.models import Q
from django.conf import settings
from django.contrib.auth import get_user_model

from rest_framework.exceptions import APIException
from rest_framework import permissions, status

from concrete_datastore.concrete.models import (
    ConcretePermission,
    DIVIDER_MODEL,
    UNDIVIDED_MODEL,
)


DIVIDER_MODELs = "{}s".format(DIVIDER_MODEL)
DIVIDER_MODELs_LOWER = DIVIDER_MODELs.lower()
DIVIDER_MODEL_LOWER = DIVIDER_MODEL.lower()


class PreconditionFailed(APIException):
    status_code = status.HTTP_412_PRECONDITION_FAILED
    default_detail = 'Precondition Fail to perform this action'
    default_code = 'precondition_failed'


CONCRETE_SETTINGS = getattr(settings, 'CONCRETE', {})
API_PERMISSIONS_CLASSES = CONCRETE_SETTINGS.get('API_PERMISSIONS_CLASSES', {})


def get_permissions_classes_by_meta_model(meta_model):
    permissions_classes_path = API_PERMISSIONS_CLASSES.get(
        meta_model._specifier.name,
        ('concrete_datastore.api.v1.permissions.UserAccessPermission',),
    )

    permissions_classes = ()

    for permission_class_path in permissions_classes_path:
        module = import_module(
            permission_class_path[: permission_class_path.rfind('.')]
        )

        try:
            permissions_classes += (
                getattr(module, permission_class_path.split('.')[-1]),
            )
        except AttributeError:
            raise RuntimeError(
                'CONCRETE impoperly configured : unknown '
                'permission class {}'.format(permission_class_path)
            )
    return permissions_classes


def check_minimum_level(method, user, model):
    authenticated = user.is_authenticated
    superuser = user.is_superuser is True
    at_least_admin = False if user.is_anonymous else user.is_at_least_admin
    staff = False if user.is_anonymous else user.is_at_least_staff
    at_least_anonymous = True  #: Always true

    METHOD_TO_LEVEL_NAME = {
        'GET': '__retrieve_minimum_level__',
        'POST': '__creation_minimum_level__',
        'PUT': '__update_minimum_level__',
        'PATCH': '__update_minimum_level__',
        'DELETE': '__delete_minimum_level__',
    }

    PERMISSION_LEVELS = {
        'superuser': superuser,
        'admin': at_least_admin,
        'staff': staff,
        'manager': staff,
        'authenticated': authenticated,
        'anonymous': at_least_anonymous,
    }

    level_name = METHOD_TO_LEVEL_NAME[method]
    level_value = getattr(model, level_name)
    user_allowed = PERMISSION_LEVELS[level_value]
    return user_allowed


def check_roles(method, user, model):
    # Checking roles doesn't applies to admin and SU
    at_least_admin = False if user.is_anonymous else user.is_at_least_admin
    if at_least_admin:
        return True

    allowed = False
    user_roles = user.get_roles()
    permission_model, _ = ConcretePermission.objects.get_or_create(
        model_name=model.__name__
    )

    model_allowed_roles_qs = {
        'GET': permission_model.retrieve_roles,
        'POST': permission_model.create_roles,
        'PUT': permission_model.update_roles,
        'PATCH': permission_model.update_roles,
        'DELETE': permission_model.delete_roles,
    }

    # Filter the queryset corresponding to the method with user's roles
    allowed = (
        model_allowed_roles_qs[method].filter(name__in=user_roles).exists()
    )

    return allowed


def does_intersect(queryset_1, queryset_2):
    if queryset_1.model is not queryset_2.model:
        raise ValueError(
            f"Attempting intersection between two querysets of different models"
        )
    return queryset_1.filter(
        pk__in=queryset_2.values_list('pk', flat=True)
    ).exists()


class UserAccessPermission(permissions.BasePermission):
    message = 'User access permission refused.'

    def check_divider_permission(self, request, obj):
        if obj.__class__.__name__ in UNDIVIDED_MODEL:
            return True
        if obj.__class__.__name__ == DIVIDER_MODEL:
            return (
                getattr(request.user, DIVIDER_MODELs_LOWER)
                .filter(pk=obj.pk)
                .exists()
            )

        if getattr(obj, DIVIDER_MODEL_LOWER) is None:
            return True
        return self.is_obj_divider_accessible_by_user(request, obj)

    def is_obj_divider_accessible_by_user(self, request, obj):
        return (
            getattr(request.user, DIVIDER_MODELs_LOWER)
            .filter(pk=getattr(obj, DIVIDER_MODEL_LOWER).pk)
            .exists()
        )

    def has_permission(self, request, view):
        if not hasattr(view, 'model_class'):
            # Validated by LCO on 2019-05-15
            # In this case, if that ever happened, view should be set to
            # view = request.parser_context["view"]
            raise AttributeError('View has no model_class attr, see comment')
        if not request.user.is_anonymous:
            if not request.user.is_confirmed():
                raise PreconditionFailed(
                    detail={
                        'message': 'Email has not been validated',
                        "_errors": ["EMAIL_NOT_VALIDATED"],
                    }
                )
        if request.method not in ["OPTIONS", "HEAD"]:
            model = view.model_class
            level_allowed = check_minimum_level(
                request.method, request.user, model
            )
            if settings.USE_CONCRETE_ROLES and model != get_user_model():
                roles_allowed = check_roles(
                    request.method, request.user, model
                )
                return level_allowed and roles_allowed
            else:
                return level_allowed

        return True

    def is_field_not_divider_changed(self, request):
        # If a key different of the divider is in the data, data will be changed
        if request.data:
            for key in request.data:
                if key != DIVIDER_MODELs_LOWER:
                    return True
        # If the divider is not in the request data
        return False

    def can_access_user_obj(self, request, user, obj):
        field_not_divider_changed = self.is_field_not_divider_changed(request)
        #: If the obj has a lower level, authorize
        if obj < user:
            return True
        #: If the obj has a greater level, deny
        if obj > user:
            return False
        # If the obj has the same level, check if no data has been changed except divider
        if field_not_divider_changed is False:
            return True
        return False

    def can_delete_user_obj(self, request, user, obj):
        superuser = user.is_superuser is True
        at_least_staff = False if user.is_anonymous else user.is_at_least_staff

        if superuser:
            return True
        # If it at least staff, check levels
        if at_least_staff:
            #: If the obj has a lower level, authorize
            if obj < user:
                return True
            #: If the obj has a greater or equal level, deny
            else:
                return False

        #: If it is not at least staff, deny
        return False

    def has_object_permission(self, request, view, obj):
        user = request.user
        superuser = user.is_superuser is True
        at_least_staff = False if user.is_anonymous else user.is_at_least_staff
        authenticated = user.is_authenticated is True
        if request.method in permissions.SAFE_METHODS:
            return (
                obj.public is True
                or isinstance(obj, get_user_model())
                or authenticated
            )

        elif isinstance(obj, get_user_model()):
            # UPDATE User only if user or superuser
            is_himself = obj.pk == user.pk
            if request.method == 'DELETE':
                return self.can_delete_user_obj(request, user, obj)
            if is_himself:
                return True
            if superuser:
                return True
            elif at_least_staff:
                return self.can_access_user_obj(request, user, obj)
            else:
                return False
        else:
            return authenticated and (
                (obj.created_by is not None and obj.created_by.pk == user.pk)
                or (user.is_at_least_admin)
                or (
                    obj.can_admin_users.filter(pk=user.pk).exists()
                    or does_intersect(
                        obj.can_admin_groups, user.concrete_groups
                    )
                )
                or (
                    at_least_staff
                    and self.check_divider_permission(request, obj) is True
                )
            )


def get_available_scope_pks_for(user):
    all_divider_pk = list(
        set(getattr(user, DIVIDER_MODELs_LOWER).values_list('pk', flat=True))
    )
    return all_divider_pk


def apply_scope_filters(user, queryset):
    model_name = queryset.model.__name__
    model_is_divided = model_name not in UNDIVIDED_MODEL
    if not model_is_divided:
        return queryset.none()

    divider_query = "{}__pk__in".format(DIVIDER_MODEL_LOWER)
    all_divider_pk = get_available_scope_pks_for(user)
    return queryset.filter(**{divider_query: all_divider_pk})


def filter_queryset_by_divider(queryset, user, divider):
    """
    queryset is the one of the current view,
    user is the connected one or anonymous
    divider is the one in headers and guarantee that current model is divided
    """
    if divider is None:
        raise ValueError("The divider should not be null")

    # On check en premier le divider
    queryset = queryset.filter(**{DIVIDER_MODEL_LOWER: divider.uid})

    if user.is_superuser is True:
        return queryset
    elif getattr(user, 'admin', False) is True:
        return queryset

    # Authenticated users
    if user.is_authenticated is True:
        queryset = apply_scope_filters(user, queryset)

    return queryset


def filter_queryset_by_permissions(queryset, user, divider):
    if queryset.model == get_user_model():
        raise ValueError(
            "Queryset of model User cannot be filtered by permissions"
        )
    all_public = Q(public=True)

    #: Anonymous user can only see public objects
    if user.is_authenticated is not True:
        return queryset.filter(all_public)

    # OBSOLETE RIGHT NOW
    # SHOULD BE DISCUSSED
    #: If current user a "local administrator" on the divided model
    # try:
    #     if getattr(
    #         user, DIVIDER_MODELs_LOWER
    #     ).filter(divider_admin=True).exists():
    #         return queryset
    # except Exception:
    #     #: Pass explicit
    #     pass

    all_owned = Q(created_by__pk=user.pk)
    all_administrables = Q(can_admin_users__pk=user.pk)
    all_viewables = Q(can_view_users__pk=user.pk)
    all_group_viewables = Q(
        can_view_groups__pk__in=user.concrete_groups.values_list(
            'pk', flat=True
        )
    )
    all_group_administrables = Q(
        can_admin_groups__pk__in=user.concrete_groups.values_list(
            'pk', flat=True
        )
    )
    filterd_queryset = queryset.filter(
        all_public
        | all_owned
        | all_administrables
        | all_viewables
        | all_group_viewables
        | all_group_administrables
    )
    #: Segment data per user level
    if user.is_superuser is True:
        return queryset
    elif getattr(user, 'admin', False) is True:
        return queryset
    elif user.is_staff is True:
        #: Objects that should be returned must have a FK (from the scoping
        #: relation) towards the scope model *instances* (formerly named divider)
        queryset = (
            apply_scope_filters(user, queryset) | filterd_queryset
        ).distinct()
        return queryset
    else:
        queryset = filterd_queryset
        #: Here we may have redundant objects, apply distinct on queryset
        #: May be unsupported by sqlite
        queryset = queryset.distinct()

    return queryset



# coding: utf-8
import json
import logging

from django.conf import settings
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, render, redirect
from django.contrib.auth import get_user_model
from concrete_datastore.concrete.models import UserConfirmation, DIVIDER_MODEL

logger = logging.getLogger(__name__)


def email_confirmation_view(request, token):
    confirmation = get_object_or_404(UserConfirmation, pk=token)
    confirmation.confirmed = True
    confirmation.save()
    if confirmation.redirect_to:
        return redirect(confirmation.redirect_to)
    return HttpResponse('Compte confirmé, Account confirmed')


def filter_scopes_for_unsubscribe(queryset, lookup_as_dict):
    return queryset.filter(**lookup_as_dict)


def parse_lookup_for_unsubscribe(model, lookup_as_json, fields):
    try:
        lookup_as_dict = json.loads(lookup_as_json)
        if not isinstance(lookup_as_dict, dict):
            raise json.decoder.JSONDecodeError(
                msg="JSON main object should be a dict",
                doc=lookup_as_dict,
                pos=0,
            )

        for k in lookup_as_dict.keys():
            if k not in fields:
                raise AttributeError(f'{k} not in {model.__name__} fields')

        return lookup_as_dict
    except json.decoder.JSONDecodeError:
        raise AttributeError(
            'Concrete settings '
            'CONCRETE_SCOPES_FILTER_LOOKUP_FOR_UNSUBSCRIBE_JSON '
            'should be a valid json'
        )


def unsubscribe_notifications_view(request, token):
    """
    token is the uid of UserNotificationChoice instance.
    """
    user = get_object_or_404(
        get_user_model(), subscription_notification_token=token
    )
    lookup_as_json = (
        settings.CONCRETE_SCOPES_FILTER_LOOKUP_FOR_UNSUBSCRIBE_JSON
    )

    queryset = getattr(user, '{}s'.format(DIVIDER_MODEL.lower())).all()
    scopes = filter_scopes_for_unsubscribe(
        queryset=queryset,
        lookup_as_dict=parse_lookup_for_unsubscribe(
            model=queryset.model,
            lookup_as_json=lookup_as_json,
            fields=[f.name for f in queryset.model._meta.fields],
        ),
    )
    return render(
        request,
        'mainApp/unsubscribe.html',
        {
            "token": token,
            "user_email": user.email,
            "scopes": scopes,
            'error_message': '',
            'platform_name': settings.PLATFORM_NAME,
        },
    )


def unsubscribe_notifications_result_view(request, token):
    error_message = ''
    try:
        user = get_object_or_404(
            get_user_model(), subscription_notification_token=token
        )
        results = request.POST.dict()
        # Accept integers (0/1)or boolean (False/True) in answer
        unsub_all = bool(results.get('all', 0))
        list_unsub_to_uid = request.POST.getlist('scope', [])

        user.unsubscribe_all = unsub_all
        user.unsubscribe_to.set(list_unsub_to_uid)
        user.save()
    except Exception:
        error_message = "Something went wrong when validating your choice"
        logger.exception(f'Unsubscription failed for user {user}')
    return render(
        request,
        'mainApp/unsubscribe_result.html',
        {
            'error_message': error_message,
            'platform_name': settings.PLATFORM_NAME,
        },
    )



# coding: utf-8
from django.conf import settings

from concrete_datastore.parsers.loaders import loads_meta


default_meta_models = [
    # Model Email
    {
        "std.name": "Email",
        "std.specifier": "Model",
        "std.verbose_name": "Email",
        "std.verbose_name_plural": "Emails",
        "std.description": "Email",
        "ext.m_creation_minimum_level": "authenticated",
        "ext.m_retrieve_minimum_level": "superuser",
        "ext.m_update_minimum_level": "superuser",
        "ext.m_delete_minimum_level": "superuser",
        "ext.m_is_default_public": False,
        "ext.m_unicode": 'subject',
        "ext.m_list_display": ['subject', 'receiver'],
        "ext.m_search_fields": ['subject', 'body'],
        "ext.m_filter_fields": [],
        "std.fields": [
            {
                "std.name": "subject",
                "std.specifier": "Field",
                "std.verbose_name": "subject",
                "std.description": "subject of the email",
                "std.type": "data",
                "ext.f_type": "CharField",
                "ext.f_args": {'max_length': 250},
            },
            {
                "std.name": "resource_status",
                "std.specifier": "Field",
                "std.verbose_name": "Resource status",
                "std.description": "status of the resource email",
                "std.type": "data",
                "ext.f_type": "CharField",
                "ext.f_args": {
                    'max_length': 250,
                    'default': 'to-send',
                    'choices': (
                        ('to-send', 'Email should be sent'),
                        ('send-in-progress', 'Email sending'),
                        ('send-error', 'Error during sending'),
                        ('sent', 'Email sent'),
                    ),
                },
            },
            {
                "std.name": "resource_message",
                "std.specifier": "Field",
                "std.verbose_name": "Resource message",
                "std.description": "message of the resource email",
                "std.type": "data",
                "ext.f_type": "CharField",
                "ext.f_args": {'max_length': 250},
            },
            {
                "std.name": "body",
                "std.specifier": "Field",
                "std.verbose_name": "body",
                "std.description": "body of the email",
                "std.type": "data",
                "ext.f_type": "TextField",
                "ext.f_args": {},
            },
            {
                "std.name": "receiver",
                "std.specifier": "Field",
                "std.verbose_name": "receiver",
                "std.description": "receiver of the email",
                "std.type": "rel_single",
                "ext.f_type": "ForeignKey",
                "ext.f_args": {
                    'to': 'concrete.User',
                    'related_name': 'received_emails',
                    'null': False,
                    'blank': False,
                },
            },
        ],
    }
]


meta_models = loads_meta(settings.META_MODEL_DEFINITIONS) + loads_meta(
    default_meta_models
)


list_of_meta = list(
    filter(
        lambda x: x._specifier.name not in settings.DISABLED_MODELS,
        meta_models,
    )
)

meta_registered = {
    k: v
    for k, v in zip(
        map(lambda m: 'concrete.{}'.format(m._specifier.name), list_of_meta),
        list_of_meta,
    )
}


def make_unicode_method(meta_model):
    def __str__(self):
        model_field_name = meta_model.get_property('m_unicode')
        try:
            return str(getattr(self, model_field_name))
        except Exception:
            return '<{} {}>'.format(meta_model.get_model_name(), self.pk)

    return __str__


def get_meta_definition_by_model_name(model_name):
    for meta_definition in list_of_meta:
        if meta_definition.get_model_name() == model_name:
            return meta_definition
    raise AttributeError('No model named {}'.format(model_name))



# coding: utf-8
import pendulum
import binascii
import logging
import uuid
import sys
import os

from collections import defaultdict
from itertools import chain
from binascii import unhexlify
from datetime import date

from django.contrib.postgres.fields import JSONField
from django.contrib.auth.models import AbstractUser, BaseUserManager
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ObjectDoesNotExist
from django.utils.encoding import force_text
from django_otp.models import Device
from django_otp.oath import totp
from django_otp.util import hex_validator, random_hex
from django.utils import timezone
from django.conf import settings
from django.urls import reverse
from django.apps import apps
from django.contrib.gis.db import models  # it includes all default fields

from rest_framework.authtoken.models import Token

from concrete_datastore.concrete.meta import (
    get_meta_definition_by_model_name,
    meta_models,
    make_unicode_method,
)
from concrete_datastore.api.v1.validators import validate_file
from concrete_datastore.concrete.constants import (
    ATLEAST_LEVEL_ATTRS,
    EXACT_LEVEL_ATTRS,
    CRUD_LEVEL,
    LIST_USER_LEVEL,
    HANDELED_MODELISATION_VERSIONS,
    TYPE_EQ,
)

# Since a lot of models are meta-declared, deactivate pylint no-member here
# pylint: disable=no-member


def get_fields_and_types_of_model(model):
    fields_and_types = defaultdict(lambda: 'Unknow type')
    meta_model = get_meta_definition_by_model_name(model.__name__)
    for field_name, field in meta_model.get_fields():
        fields_and_types[field_name] = TYPE_EQ.get(field.f_type, 'Unknow type')
    return fields_and_types


def compute_auth_token_expiry():
    now = pendulum.now('utc')
    if settings.API_TOKEN_EXPIRY == 0:
        return now.add(years=100)
    return now.add(minutes=settings.API_TOKEN_EXPIRY)


def compute_pwd_change_token_expiry():
    now = pendulum.now('utc')
    return now.add(hours=settings.PASSWORD_CHANGE_TOKEN_EXPIRY_HOURS)


class AuthToken(Token):
    user = models.ForeignKey(
        'concrete.User',
        related_name='auth_tokens',
        on_delete=models.CASCADE,
        verbose_name=_("User"),
    )
    expired = models.BooleanField(default=False)
    expiration_date = models.DateTimeField(default=compute_auth_token_expiry)
    last_action_date = models.DateTimeField(default=timezone.now)


def default_key():
    return force_text(random_hex(20))


def key_validator(value):
    return hex_validator()(value)


class EmailDevice(Device):
    """
    A :class:`~django_otp.models.Device` that delivers a token to the user's
    registered email address (``user.email``). This is intended for
    demonstration purposes; if you allow users to reset their passwords via
    email, then this provides no security benefits.

    .. attribute:: key

        *CharField*: A hex-encoded secret key of up to 40 bytes. (Default: 20
        random bytes)
    """

    uid = models.UUIDField(default=uuid.uuid4, primary_key=True)
    key = models.CharField(
        max_length=80,
        validators=[key_validator],
        default=default_key,
        help_text='A hex-encoded secret key of up to 20 bytes.',
    )
    email = models.CharField(
        max_length=250,
        default='',
        help_text='Email address to send verification code.',
    )
    modification_date = models.DateTimeField(auto_now=True)
    creation_date = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(
        'concrete.User',
        related_name="owned_%(class)ss",
        null=True,
        on_delete=models.PROTECT,
    )

    @property
    def id(self):
        return self.pk

    @property
    def bin_key(self):
        return unhexlify(self.key.encode())

    def generate_challenge(self):
        code_timeout = settings.TWO_FACTOR_CODE_TIMEOUT_SECONDS
        token = totp(self.bin_key, step=code_timeout)

        main_app = apps.get_app_config('concrete')
        body = settings.TWO_FACTOR_TOKEN_MSG.format(
            platform_name=settings.PLATFORM_NAME,
            confirm_code=token,
            min_validity=int(code_timeout / 60),
        )
        main_app.models['email'].objects.create(
            receiver=self.user,
            body=body,
            resource_status='to-send',
            subject='[Confirmation authentication {}]'.format(
                settings.PLATFORM_NAME
            ),
            created_by=self.user,
        )
        logging.debug(
            'Confirm Email has been sent to {}'.format(self.user.email)
        )

        return token

    def verify_token(self, token):
        try:
            token = int(token)
        except Exception:
            verified = False
        else:
            verified = any(
                totp(
                    self.bin_key,
                    step=settings.TWO_FACTOR_CODE_TIMEOUT_SECONDS,
                    drift=drift,
                )
                == token
                for drift in [0, -1]
            )

        return verified

    @classmethod
    def from_persistent_id(cls, persistent_id):
        """
        Loads a device from its persistent id::

            device == Device.from_persistent_id(device.persistent_id)

        """
        device = None

        try:
            model_label, device_id = persistent_id.rsplit('/', 1)
            app_label, model_name = model_label.split('.')

            main_app = apps.get_app_config(app_label)

            device_cls = main_app.models[model_name]
            if issubclass(device_cls, Device):
                device = device_cls.objects.filter(pk=device_id).first()
        except (ValueError, LookupError):
            pass

        return device


class TemporaryToken(models.Model):
    """
    The temporary authorization token model for two factor auth
    """

    key = models.CharField(_("Key"), max_length=40, primary_key=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name='temporary_tokens',
        on_delete=models.CASCADE,
        verbose_name=_("User"),
    )
    expired = models.BooleanField(default=False)
    creation_date = models.DateTimeField(auto_now_add=True)
    modification_date = models.DateTimeField(auto_now=True)
    ordering = ('-modification_date', '-creation_date')

    class Meta:
        verbose_name = _("Temporary Token")
        verbose_name_plural = _("Temporary Tokens")

    def save(self, *args, **kwargs):
        if not self.key:
            self.key = self.generate_key()
        return super().save(*args, **kwargs)

    def generate_key(self):
        return binascii.hexlify(os.urandom(20)).decode()

    def __str__(self):
        return self.key


class ConcreteRole(models.Model):
    class Meta:
        verbose_name = 'Role'

    uid = models.UUIDField(default=uuid.uuid4, primary_key=True)
    name = models.CharField(
        unique=True, blank=False, null=False, max_length=255
    )
    users = models.ManyToManyField(
        'concrete.User', blank=True, related_name='concrete_roles'
    )
    created_by = models.ForeignKey(
        'concrete.User',
        related_name="owned_%(class)ss",
        null=True,
        on_delete=models.PROTECT,
    )
    modification_date = models.DateTimeField(
        auto_now=True,
        # default=timezone.now
    )
    creation_date = models.DateTimeField(
        auto_now_add=True,
        # default=timezone.now
    )

    def __str__(self):
        # the __str__ method should return a string
        return str(self.name)


class ConcretePermission(models.Model):
    class Meta:
        verbose_name = 'Permission'

    uid = models.UUIDField(default=uuid.uuid4, primary_key=True)
    model_name = models.CharField(
        unique=True, blank=False, null=False, max_length=255
    )

    create_roles = models.ManyToManyField(
        ConcreteRole, blank=True, related_name='create_permissions'
    )
    retrieve_roles = models.ManyToManyField(
        ConcreteRole, blank=True, related_name='retrieve_permissions'
    )
    update_roles = models.ManyToManyField(
        ConcreteRole, blank=True, related_name='update_permissions'
    )
    delete_roles = models.ManyToManyField(
        ConcreteRole, blank=True, related_name='delete_permissions'
    )
    created_by = models.ForeignKey(
        'concrete.User',
        related_name="owned_%(class)ss",
        null=True,
        on_delete=models.PROTECT,
    )

    modification_date = models.DateTimeField(auto_now=True)
    creation_date = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        # the __str__ method should return a string
        return str(self.model_name)


class SecureConnectToken(models.Model):
    value = models.UUIDField(default=uuid.uuid4, primary_key=True)
    user = models.ForeignKey(
        'concrete.User',
        on_delete=models.PROTECT,
        related_name='secure_connect_tokens',
    )
    creation_date = models.DateTimeField(auto_now_add=True)
    modification_date = models.DateTimeField(auto_now=True)
    expired = models.BooleanField(default=False)

    mail_sent = models.BooleanField(default=False)
    url = models.URLField(blank=True, null=True)

    def __str__(self):
        return str(self.value)

    def send_mail(self):
        main_app = apps.get_app_config('concrete')
        body = settings.SECURE_TOKEN_MESSAGE_BODY.format(
            platform=settings.PLATFORM_NAME,
            email=self.user.email,
            link=self.url,
        )

        main_app.models['email'].objects.create(
            receiver=self.user,
            body=body,
            resource_status='to-send',
            subject='[Authentication {}]'.format(settings.PLATFORM_NAME),
            created_by=self.user,
        )
        logging.debug(
            'Confirm Email has been sent to {}'.format(self.user.email)
        )
        self.mail_sent = True
        self.save()


class UserConfirmation(models.Model):
    uid = models.UUIDField(default=uuid.uuid4, primary_key=True)
    confirmed = models.BooleanField(default=False)
    user = models.ForeignKey(
        'concrete.User', on_delete=models.CASCADE, related_name='confirmations'
    )
    creation_date = models.DateTimeField(auto_now_add=True)
    modification_date = models.DateTimeField(auto_now=True)
    redirect_to = models.URLField(null=True)
    link_sent = models.BooleanField(default=False)

    @property
    def url(self):
        return '{scheme}://{domain}:{port}{uri}'.format(
            scheme=settings.SCHEME,
            domain=settings.HOSTNAME,
            port=settings.PORT,
            uri=reverse(
                'concrete:email_confirmation', kwargs={'token': self.uid}
            ),
        )

    def send_link(self, body=None):
        main_app = apps.get_app_config('concrete')
        if body is None:
            body = settings.AUTH_CONFIRM_EMAIL_MESSAGE_BODY.format(
                platform=settings.PLATFORM_NAME,
                email=self.user.email,
                link=self.url,
            )
        main_app.models['email'].objects.create(
            receiver=self.user,
            body=body,
            resource_status='to-send',
            subject='Confirmation mail',
            created_by=self.user,
        )
        logging.debug(
            'Confirm Email has been sent to {}'.format(self.user.email)
        )
        self.link_sent = True
        self.save()


class ConfirmableUserAbstract(models.Model):
    class Meta:
        abstract = True

    def is_confirmed(self):
        if settings.AUTH_CONFIRM_EMAIL_ENABLE is False:
            return True
        confirmation = self.confirmations
        try:
            return confirmation.latest('modification_date').confirmed
        except ObjectDoesNotExist:
            return False

    def get_or_create_confirmation(self, redirect_to=None):
        if redirect_to is None:
            redirect_to = settings.AUTH_CONFIRM_EMAIL_DEFAULT_REDIRECT_TO
        try:
            return self.confirmations.latest('modification_date')
        except ObjectDoesNotExist:
            return self.confirmations.create(redirect_to=redirect_to)

    def confirmate(self):
        self.confirmations.all().update(confirmed=True)


class HasPermissionAbstractUser(models.Model):
    _is_verified = False

    def is_verified(self):
        return self._is_verified

    class Meta:
        abstract = True

    def __setattr__(self, attrname, val):
        if attrname == 'level':
            kwargs = {'level': val, 'commit': True}
            getattr(self, 'set_level')(**kwargs)
        else:
            super(HasPermissionAbstractUser, self).__setattr__(attrname, val)

    @property
    def is_at_least_admin(self):
        return self.admin is True or self.is_superuser is True

    @property
    def is_at_least_staff(self):
        return self.is_staff is True or self.is_at_least_admin is True

    def set_level(self, level, commit=False):
        list_levels = LIST_USER_LEVEL

        if type(level) != str:
            level = str(level)

        level = level.lower()

        if level in list_levels:
            self.is_superuser = level in ('superuser',)
            self.admin = level in ('admin', 'superuser')
            self.is_staff = level in ('superuser', 'admin', 'manager')
            self.is_active = level != 'blocked'

        if commit is True:
            self.save()

    def get_level(self):
        if self.is_active is False:
            return 'Blocked'
        elif self.is_superuser:
            return 'SuperUser'
        elif self.admin:
            return 'Admin'
        elif self.is_staff:
            return 'Manager'
        return 'SimpleUser'

    def get_roles(self):
        roles = self.concrete_roles.values_list('name', flat=True)
        return roles

    def get_roles_uid(self):
        roles = self.concrete_roles.values_list('uid', flat=True)
        return roles

    @property
    def id(self):
        return self.pk

    @property
    def level(self):
        if self.is_superuser:
            return 'superuser'
        elif self.admin:
            return 'admin'
        elif self.is_staff:
            return 'manager'
        elif self.is_active:
            return 'simpleuser'
        else:
            return 'blocked'

    @classmethod
    def filter_by_at_least_level(cls, level, queryset=None):
        queryset = queryset or cls.objects.all()
        if level == 'blocked' or level not in ATLEAST_LEVEL_ATTRS.keys():
            # A user cannot be at least blocked
            return queryset.none()
        return queryset.filter(**ATLEAST_LEVEL_ATTRS[level])

    @classmethod
    def filter_by_exact_level(cls, level, queryset=None):
        queryset = queryset or cls.objects.all()
        if level not in EXACT_LEVEL_ATTRS.keys():
            return queryset.none()
        return queryset.filter(**EXACT_LEVEL_ATTRS[level])

    def __gt__(self, other):
        if isinstance(other, self.__class__) is False:
            raise ValueError(
                "The element to compare with should be an instance of User"
            )
        if LIST_USER_LEVEL.index(self.level) > LIST_USER_LEVEL.index(
            other.level
        ):
            return True
        return False

    def __ge__(self, other):
        if isinstance(other, self.__class__) is False:
            raise ValueError(
                "The element to compare with should be an instance of User"
            )
        if LIST_USER_LEVEL.index(self.level) >= LIST_USER_LEVEL.index(
            other.level
        ):
            return True
        return False

    def __lt__(self, user):
        if isinstance(user, self.__class__) is False:
            raise ValueError(
                "The element to compare with should be an instance of User"
            )
        if LIST_USER_LEVEL.index(self.level) < LIST_USER_LEVEL.index(
            user.level
        ):
            return True
        return False

    def __le__(self, user):
        if isinstance(user, self.__class__) is False:
            raise ValueError(
                "The element to compare with should be an instance of User"
            )
        if LIST_USER_LEVEL.index(self.level) <= LIST_USER_LEVEL.index(
            user.level
        ):
            return True
        return False

    @property
    def password_has_expired(self):
        expiry = settings.PASSWORD_EXPIRY_TIME
        password_has_expiry = expiry > 0
        now = pendulum.now()
        d = self.password_modification_date
        last_password_modification = pendulum.datetime(d.year, d.month, d.day)

        return (
            password_has_expiry
            and now.diff(last_password_modification).in_days() >= expiry
        )


class PasswordChangeToken(models.Model):

    uid = models.UUIDField(primary_key=True, default=uuid.uuid4)
    creation_date = models.DateTimeField(auto_now_add=True)
    expiry_date = models.DateTimeField(default=compute_pwd_change_token_expiry)
    user = models.ForeignKey(
        'concrete.User',
        related_name="reset_password_tokens",
        on_delete=models.PROTECT,
    )


class DefaultDivider(models.Model):
    uid = models.UUIDField(primary_key=True, default=uuid.uuid4)
    name = models.CharField(max_length=250, default="Default Divider")


class DeletedModel(models.Model):
    uid = models.UUIDField()

    model_name = models.CharField(max_length=255, default='')

    modification_date = models.DateTimeField(auto_now=True)

    creation_date = models.DateTimeField(auto_now_add=True)


class UserManager(BaseUserManager):
    """Define a model manager for User model with no username field."""

    use_in_migrations = True

    def _create_user(self, email, password, **extra_fields):
        """Create and save a User with the given email and password."""
        if not email:
            raise ValueError('The given email must be set')
        email = self.normalize_email(email).lower()
        user = self.model(email=email, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email, password=None, **extra_fields):
        """Create and save a regular User with the given email and password."""
        extra_fields.setdefault('is_staff', False)
        extra_fields.setdefault('is_superuser', False)
        return self._create_user(email, password, **extra_fields)

    def create_superuser(self, email, password, **extra_fields):
        """Create and save a SuperUser with the given email and password."""
        extra_fields.setdefault('is_staff', True)
        extra_fields.setdefault('is_superuser', True)

        if extra_fields.get('is_staff') is not True:
            raise ValueError('Superuser must have is_staff=True.')
        if extra_fields.get('is_superuser') is not True:
            raise ValueError('Superuser must have is_superuser=True.')

        return self._create_user(email, password, **extra_fields)


def get_common_fields(public_default_value=False):
    if public_default_value not in (True, False):
        raise ValueError(
            "Value of 'public_default_value' should be whether True or False"
        )
    return {
        'uid': models.UUIDField(primary_key=True, default=uuid.uuid4),
        'modification_date': models.DateTimeField(
            auto_now=True,
            # default=timezone.now
        ),
        'creation_date': models.DateTimeField(
            auto_now_add=True,
            # default=timezone.now
        ),
        'public': models.BooleanField(default=public_default_value),
    }


def get_user_tracked_fields():
    return {
        'created_by': models.ForeignKey(
            'concrete.User',
            related_name="owned_%(class)ss",
            null=True,
            on_delete=models.PROTECT,
        ),
        'can_admin_users': models.ManyToManyField(
            'concrete.User',
            related_name="administrable_%(class)ss",
            blank=True,
        ),
        'can_view_users': models.ManyToManyField(
            'concrete.User', related_name="viewable_%(class)ss", blank=True
        ),
        'can_admin_groups': models.ManyToManyField(
            'concrete.Group',
            related_name="group_administrable_%(class)ss",
            blank=True,
        ),
        'can_view_groups': models.ManyToManyField(
            'concrete.Group',
            related_name="group_viewable_%(class)ss",
            blank=True,
        ),
    }


def get_divider_fields_foreignkey(model):
    return {
        '{}'.format(DIVIDER_MODEL.lower()): models.ForeignKey(
            'concrete.{}'.format(model),
            related_name="divider_%(class)ss",
            null=True,
            blank=True,
            on_delete=models.PROTECT,
        ),
        "additional_filtering": models.BooleanField(default=False),
    }


def get_divider_fields_manytomany(model):
    return {
        '{}s'.format(DIVIDER_MODEL.lower()): models.ManyToManyField(
            'concrete.{}'.format(model),
            related_name="divider_%(class)ss",
            blank=True,
        )
    }


def get_divider_notification_fields(model):
    return {
        'unsubscribe_all': models.BooleanField(default=False),
        'unsubscribe_to': models.ManyToManyField(
            'concrete.{}'.format(model),
            related_name="unsubscribed_%(class)ss",
            blank=True,
        ),
    }


def get_minimum_level(meta_model, prop_name, default_value):
    level = meta_model.get_property(
        prop_name=prop_name, default_value=default_value
    )
    if level not in CRUD_LEVEL:
        return default_value
    return level


def make_django_model(meta_model, divider):
    class Meta:
        verbose_name = _(meta_model.get_verbose_name())
        verbose_name_plural = _(meta_model.get_verbose_name_plural())
        ordering = ('-modification_date', '-creation_date')
        unique_together = tuple(
            meta_model.get_property(
                prop_name='m_unique_together', default_value=[]
            )
        )

    modelisation_version = getattr(meta_model, 'version', None)

    if (
        modelisation_version is not None
        and modelisation_version not in HANDELED_MODELISATION_VERSIONS
    ):
        raise ValueError('Unknown modelisation format')

    creation_level = get_minimum_level(
        meta_model=meta_model,
        prop_name='m_creation_minimum_level',
        default_value='authenticated',
    )

    retrieve_level = get_minimum_level(
        meta_model=meta_model,
        prop_name='m_retrieve_minimum_level',
        default_value='authenticated',
    )

    update_level = get_minimum_level(
        meta_model=meta_model,
        prop_name='m_update_minimum_level',
        default_value='authenticated',
    )

    delete_level = get_minimum_level(
        meta_model=meta_model,
        prop_name='m_delete_minimum_level',
        default_value='superuser',
    )

    attrs = {
        'Meta': Meta,
        '__module__': 'concrete_datastore.concrete.models',
        '__str__': make_unicode_method(meta_model),
        '__creation_minimum_level__': creation_level,
        '__retrieve_minimum_level__': retrieve_level,
        '__update_minimum_level__': update_level,
        '__delete_minimum_level__': delete_level,
    }

    is_default_public = meta_model.get_property(
        prop_name='m_is_default_public', default_value=False
    )
    if is_default_public is None or is_default_public not in (True, False):
        is_default_public = False
    attrs.update(get_common_fields(public_default_value=is_default_public))

    ancestors = (models.Model,)
    if meta_model.get_model_name() == 'User':
        ancestors = (
            AbstractUser,
            HasPermissionAbstractUser,
            ConfirmableUserAbstract,
        )
        attrs.update(
            {
                'username': None,
                'objects': UserManager(),
                'admin': models.BooleanField(default=False),
                'password_modification_date': models.DateField(
                    default=date.today
                ),
                'subscription_notification_token': models.UUIDField(
                    default=uuid.uuid4, editable=False
                ),
                'login_counter': models.IntegerField(default=0),
                'external_auth': models.BooleanField(default=False),
                # True if user created by an external auth
                'USERNAME_FIELD': 'email',
                'REQUIRED_FIELDS': [],
            }
        )
    else:
        attrs.update(get_user_tracked_fields())

    for field_name, field in meta_model.get_fields():
        if field_name in attrs:
            raise ValueError(
                f'{field_name} is a protected field and cannot be overwritten'
            )

        # obsoletes fields
        if meta_model.get_model_name() == 'User':
            obsoletes_level = ('guest', 'manager', 'blocked')
            if field_name in obsoletes_level:
                raise ValueError(
                    f'The fields {obsoletes_level} are no longer supported'
                )

        args = field.f_args
        if field.f_type == 'FileField':
            args.update(
                {'blank': True, 'null': True, 'validators': [validate_file]}
            )
        elif field.f_type == 'JSONField':
            json_args = args
            json_args['blank'] = True
            json_args['encoder'] = None
            json_args['null'] = False
            json_args['default'] = dict
            attrs.update({field_name: JSONField(**json_args)})
            continue
        elif field.f_type in ('CharField', 'TextField'):
            args['null'] = False
            args.setdefault('blank', True)
            args.setdefault('default', "")
        elif field.f_type in ('IntegerField', 'BigIntegerField'):
            args['null'] = False
            args.setdefault('blank', True)
            args.setdefault('default', 0)
        elif field.f_type == 'DecimalField':
            args['null'] = False
            args.setdefault('decimal_places', 2)
            args.setdefault('max_digits', 20)
            args.setdefault('default', 0.00)
        elif field.f_type in ('ForeignKey',):

            # Copy args to not alter the real field.f_args
            args = args.copy()
            # Force FK to null=True to avoid default value problems
            args.update({'null': True})
            # PROTECT foreign key deletion by default
            on_delete_rule = args.get('on_delete', 'PROTECT')
            if on_delete_rule not in ('CASCADE', 'SET_NULL', 'PROTECT'):
                raise ValueError(
                    f'On delete rule "{on_delete_rule}" is invalid. '
                    'It must be "CASCADE", "SET_NULL" or "PROTECT"'
                )

            args.update({'on_delete': getattr(models, on_delete_rule)})

        elif field.f_type in ('GenericIPAddressField',):
            #: If blank is True, null should be too
            #: https://docs.djangoproject.com/fr/3.1/ref/models/fields/#genericipaddressfield
            if args.get('blank', False) is True:
                args['null'] = True
            else:
                args.setdefault('blank', False)
                args['null'] = False

        elif field.f_type in ('DateTimeField',):
            if args.get('null', False) is True:
                args['null'] = True
                args['blank'] = True
            else:
                args['default'] = timezone.now
                args['null'] = False
                args.setdefault('blank', False)
        elif field.f_type in ('DateField',):
            if args.get('null', False) is True:
                args['null'] = True
                args['blank'] = True
            else:
                args['default'] = date.today
                args['null'] = False
                args.setdefault('blank', False)
        elif field.f_type == 'ManyToManyField':
            # Copy args to not alter the real field.f_args
            args = args.copy()
            args.pop('null', None)
        attrs.update({field_name: getattr(models, field.f_type)(**args)})
    if meta_model.get_model_name() != divider:
        if meta_model.get_model_name() == "User":
            attrs.update(get_divider_fields_manytomany(divider))
            attrs.update(get_divider_notification_fields(divider))
            attrs.update(
                {'email': models.EmailField(_('email address'), unique=True)}
            )
        elif meta_model.get_model_name() in UNDIVIDED_MODEL:
            pass
        else:
            attrs.update(get_divider_fields_foreignkey(divider))

    return type(meta_model.get_model_name(), ancestors, attrs)


def get_divider_v0(meta_model):
    if meta_model.get_model_name() == "EntityDividerModel":
        try:
            return meta_model.get_property(prop_name="fields")[0].name
        except IndexError:
            return 'DefaultDivider'


def get_divider_v1(meta_model):
    if len(meta_model.get_property('lookups', default_value=[])) != 0:
        return meta_model.get_property('lookups')[0]['model_name']


def get_undivided_models_v0(meta_model):
    if meta_model.get_model_name() == "UndividedModel":
        try:
            return meta_model.get_property(prop_name="fields")[0].name.split(
                "_"
            )
        except IndexError:
            return []
    return []


def get_undivided_models_v1(meta_model):
    model_lookups = meta_model.get_property(
        prop_name='lookups', default_value=[]
    )
    if len(model_lookups) == 0:
        return [meta_model.get_model_name()]
    return []


divider_version_eq = {
    '0.0.0': {
        'get_divider_model': get_divider_v0,
        'get_undivided_models': get_undivided_models_v0,
    },
    '1.0.0': {
        'get_divider_model': get_divider_v1,
        'get_undivided_models': get_undivided_models_v1,
    },
}


def get_divider_fn(meta_model):
    #:  Get the divider function according to datamodel version and apply
    #:  on the meta_model to get the divider.
    version = getattr(meta_model, 'version', '0.0.0')
    fn = divider_version_eq[version]['get_divider_model']
    return fn(meta_model)


def get_divider():

    dividers_set = set(map(get_divider_fn, meta_models))

    #:  Remove None values from the set
    dividers_set.discard(None)

    if len(dividers_set) > 1:
        raise NotImplementedError(
            'Multiple dividers are not yet handeled by Concrete'
        )

    if len(dividers_set) == 0:
        return "DefaultDivider"

    return dividers_set.pop()


def get_undivided_models_fn(meta_model):
    #:  Get the undivided model function according to datamodel version and
    #:  apply on the meta_models to get the undivided models.
    version = getattr(meta_model, 'version', '0.0.0')
    fn = divider_version_eq[version]['get_undivided_models']
    return fn(meta_model)


def get_undivided_models():
    undivided_models = map(get_undivided_models_fn, meta_models)
    undivided_models_set = set(chain.from_iterable(undivided_models))

    return list(undivided_models_set)


divider = get_divider()
divider_field_name = divider.lower()

DIVIDER_MODEL = divider

UNDIVIDED_MODEL = get_undivided_models()

for meta_model in meta_models:
    if meta_model.get_model_name() in ["EntityDividerModel", "UndividedModel"]:
        logging.debug("Horrible patch for handling input specification")
        continue
    setattr(
        sys.modules[__name__],
        meta_model.get_model_name(),
        make_django_model(meta_model, divider),
    )



# coding: utf-8
from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

from django_otp.middleware import OTPMiddleware

from concrete_datastore.concrete.models import EmailDevice


class OTPCustomMiddleware(OTPMiddleware):
    def _device_from_persistent_id(self, persistent_id):
        if persistent_id.count('.') > 1:
            parts = persistent_id.split('.')
            persistent_id = '.'.join((parts[-3], parts[-1]))

        device = EmailDevice.from_persistent_id(persistent_id)

        return device



# coding: utf-8
import hashlib
import logging
from collections import defaultdict

from django.utils.translation import ugettext_lazy as _
from django.apps import AppConfig
from django.apps import apps


logger_archive_users = logging.getLogger('archive-concrete-users')


def email_to_username(email):  # nosec
    return hashlib.md5(email.lower().encode('utf-8')).hexdigest()[:30]


def archive_legacy(user, email_list):
    base_email = "ghost{}@netsach.org"
    incrementation = 0
    while base_email.format(incrementation or '') in email_list:
        incrementation += 1
    new_email = base_email.format(incrementation or '')
    previous_email = user.email
    user.email = new_email
    user.save()
    logger_archive_users.info(
        f'User {user.uid}: changed email form '
        f'{previous_email} to {new_email}'
    )
    email_list.append(new_email)


def purge_users_with_wrong_usernames(user_model):
    correspondances = defaultdict(set)
    for u in user_model.objects.all():
        correspondances[u.email].add(u)

    multi_correspondances = {
        email: users_list
        for email, users_list in correspondances.items()
        if len(users_list) > 1
    }

    email_list = list(correspondances.keys())
    if len(email_list) == 0:
        logger_archive_users.debug('No migration required for legacy users')
        return
    for email, users in multi_correspondances.items():
        for user in users:
            if user.username != email_to_username(email):
                archive_legacy(user=user, email_list=email_list)


class Config(AppConfig):
    name = 'concrete_datastore.concrete'
    label = 'concrete'
    verbose_name = _('NS concrete')

    def ready(self):
        user_model = apps.get_model('concrete', 'User')
        try:
            purge_users_with_wrong_usernames(user_model=user_model)
        except Exception:
            #: legacy process when upgrading from concrete-server
            logger_archive_users.debug(
                'App Concrete not yet loaded, skipping User checks ...'
            )

        from .automation import signal_processor

        if signal_processor is None:
            raise ValueError("Missing module 'signal_processor' in automation")



# coding: utf-8
from slugify import slugify

from concrete_datastore.parsers.helpers import camel_case_to_dash_case


class Model:
    """
    Model base class
    """

    _specifier = None

    @classmethod
    def get_verbose_name(cls):
        return cls._specifier.std_verbose_name

    @classmethod
    def get_verbose_name_plural(cls):
        verbose_name_plural = cls.get_property('verbose_name_plural')
        if verbose_name_plural is None:
            verbose_name_plural = '{}s'.format(cls.get_verbose_name())
        return verbose_name_plural

    @classmethod
    def get_description(cls):
        return cls._specifier.std_description

    @classmethod
    def get_slugified_name(cls):
        return slugify(cls.get_model_name())

    @classmethod
    def get_model_name(cls):
        return cls._specifier.std_name

    @classmethod
    def get_dashed_case_class_name(cls):
        return camel_case_to_dash_case(cls.get_model_name())

    @classmethod
    def get_fields(cls):
        """
        Return a generator yielding field names
        """
        return ((f.std_name, f) for f in cls._specifier.std_fields)

    @classmethod
    def get_property(cls, prop_name, default_value=None):
        """
        Return a property of the model
        """
        return getattr(cls._specifier, prop_name, default_value)


class Model_V1:
    """
    Model base class
    """

    _specifier = None

    version = "1.0.0"

    @classmethod
    def get_verbose_name(cls):
        return getattr(
            cls._specifier, 'std_verbose_name', cls.get_model_name()
        )

    @classmethod
    def get_verbose_name_plural(cls):
        verbose_name_plural = cls.get_property('verbose_name_plural')
        if verbose_name_plural is None:
            verbose_name_plural = '{}s'.format(cls.get_verbose_name())
        return verbose_name_plural

    @classmethod
    def get_description(cls):
        return getattr(cls._specifier, 'std_description', '')

    @classmethod
    def get_slugified_name(cls):
        return slugify(cls.get_model_name())

    @classmethod
    def get_model_name(cls):
        return cls._specifier.name

    @classmethod
    def get_dashed_case_class_name(cls):
        return camel_case_to_dash_case(cls.get_model_name())

    @classmethod
    def get_fields(cls):
        """
        Return a generator yielding field names
        """
        return ((f.name, f) for f in cls._specifier.fields_spec)

    @classmethod
    def get_property(cls, prop_name, default_value=None):
        """
        Return a property of the model
        """
        return getattr(cls._specifier, prop_name, default_value)



# coding: utf-8
from importlib import import_module

from django.conf import settings

from concrete_datastore.settings.celery import app


@app.task
def async_run_plugin_tasks():
    if isinstance(settings.PLUGINS_TASKS_FUNC, dict) is False:
        return
    for path_task, queue in settings.PLUGINS_TASKS_FUNC.items():
        module_name, func_name = path_task.rsplit('.', 1)
        module = import_module(module_name)
        function = getattr(module, func_name)
        function.apply_async(queue=queue)
